from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, List, Tuple, Union

from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import YulAssignment, YulFunctionCall, YulIdentifier, YulLiteral

from .abc import YulAbc, YulStatementAbc
from .function_call import FunctionCall
from .identifier import Identifier
from .literal import Literal

if TYPE_CHECKING:
    from .block import Block


class Assignment(YulStatementAbc):
    """
    TBD
    """

    _parent: Block
    _value: Union[FunctionCall, Identifier, Literal]
    _variable_names: List[Identifier]

    def __init__(self, init: IrInitTuple, assignment: YulAssignment, parent: YulAbc):
        super().__init__(init, assignment, parent)
        if isinstance(assignment.value, YulFunctionCall):
            self._value = FunctionCall(init, assignment.value, self)
        elif isinstance(assignment.value, YulIdentifier):
            self._value = Identifier(init, assignment.value, self)
        elif isinstance(assignment.value, YulLiteral):
            self._value = Literal(init, assignment.value, self)
        else:
            assert False, f"Unexpected type: {type(assignment.value)}"
        self._variable_names = [
            Identifier(init, variable_name, self)
            for variable_name in assignment.variable_names
        ]

    def __iter__(self) -> Iterator[YulAbc]:
        yield self
        yield from self._value
        for variable_name in self._variable_names:
            yield from variable_name

    @property
    def parent(self) -> Block:
        return self._parent

    @property
    def value(self) -> Union[FunctionCall, Identifier, Literal]:
        return self._value

    @property
    def variable_names(self) -> Tuple[Identifier, ...]:
        return tuple(self._variable_names)



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Union

from ...nodes import YulIdentifier
from ..utils import IrInitTuple
from .abc import YulAbc

if TYPE_CHECKING:
    from ..statement.inline_assembly import ExternalReference
    from .assignment import Assignment
    from .expression_statement import ExpressionStatement
    from .for_loop import ForLoop
    from .function_call import FunctionCall
    from .if_statement import If
    from .switch import Switch
    from .variable_declaration import VariableDeclaration


class Identifier(YulAbc):
    """
    TBD
    """

    _parent: Union[
        Assignment,
        ExpressionStatement,
        ForLoop,
        If,
        Switch,
        VariableDeclaration,
        FunctionCall,
    ]
    _name: str
    _external_reference: Optional[ExternalReference]

    def __init__(self, init: IrInitTuple, identifier: YulIdentifier, parent: YulAbc):
        super().__init__(init, identifier, parent)
        self._name = identifier.name
        self._external_reference = None

    @property
    def parent(
        self,
    ) -> Union[
        Assignment,
        ExpressionStatement,
        ForLoop,
        If,
        Switch,
        VariableDeclaration,
        FunctionCall,
    ]:
        return self._parent

    @property
    def name(self) -> str:
        return self._name

    @property
    def external_reference(self) -> Optional[ExternalReference]:
        return self._external_reference



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING, Iterator, Optional, Set, Tuple, Union

from woke.ast.enums import ModifiesStateFlag
from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.ir.expression.abc import ExpressionAbc
from woke.ast.ir.statement.abc import StatementAbc
from woke.ast.ir.statement.expression_statement import ExpressionStatement
from woke.ast.ir.statement.variable_declaration_statement import (
    VariableDeclarationStatement,
)
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import (
    SolcExpressionStatement,
    SolcForStatement,
    SolcVariableDeclarationStatement,
)

if TYPE_CHECKING:
    from .block import Block
    from .do_while_statement import DoWhileStatement
    from .if_statement import IfStatement
    from .unchecked_block import UncheckedBlock
    from .while_statement import WhileStatement


class ForStatement(StatementAbc):
    """
    !!! example
        Lines 2-4 in the following code:
        ```solidity linenums="1"
        function foo(uint x) public pure {
            for (uint i = 0; i < 10; i++) {
                x += 1;
            }
        }
        ```
    """

    _ast_node: SolcForStatement
    _parent: Union[
        Block,
        DoWhileStatement,
        ForStatement,
        IfStatement,
        UncheckedBlock,
        WhileStatement,
    ]

    _body: StatementAbc
    _condition: Optional[ExpressionAbc]
    _initialization_expression: Optional[
        Union[ExpressionStatement, VariableDeclarationStatement]
    ]
    _loop_expression: Optional[ExpressionStatement]

    def __init__(self, init: IrInitTuple, for_: SolcForStatement, parent: SolidityAbc):
        super().__init__(init, for_, parent)
        self._body = StatementAbc.from_ast(init, for_.body, self)

        self._condition = (
            ExpressionAbc.from_ast(init, for_.condition, self)
            if for_.condition
            else None
        )

        if for_.initialization_expression is None:
            self._initialization_expression = None
        else:
            if isinstance(for_.initialization_expression, SolcExpressionStatement):
                self._initialization_expression = ExpressionStatement(
                    init, for_.initialization_expression, self
                )
            elif isinstance(
                for_.initialization_expression, SolcVariableDeclarationStatement
            ):
                self._initialization_expression = VariableDeclarationStatement(
                    init, for_.initialization_expression, self
                )

        self._loop_expression = (
            ExpressionStatement(init, for_.loop_expression, self)
            if for_.loop_expression
            else None
        )

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        yield from self._body
        if self._condition is not None:
            yield from self._condition
        if self._initialization_expression is not None:
            yield from self._initialization_expression
        if self._loop_expression is not None:
            yield from self._loop_expression

    @property
    def parent(
        self,
    ) -> Union[
        Block,
        DoWhileStatement,
        ForStatement,
        IfStatement,
        UncheckedBlock,
        WhileStatement,
    ]:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def body(self) -> StatementAbc:
        """
        Returns:
            Body of the for loop.
        """
        return self._body

    @property
    def condition(self) -> Optional[ExpressionAbc]:
        """
        !!! example
            ```solidity
            i < 10
            ```
            in the following for loop:
            ```solidity
            for (uint i = 0; i < 10; i++) {}
            ```
        Returns:
            Condition of the for loop, if any.
        """
        return self._condition

    @property
    def initialization_expression(
        self,
    ) -> Optional[Union[ExpressionStatement, VariableDeclarationStatement]]:
        """
        !!! example
            ```solidity
            uint i = 0
            ```
            in the following for loop:
            ```solidity
            for (uint i = 0; i < 10; i++) {}
            ```
        Returns:
            Initialization expression of the for loop, if any.
        """
        return self._initialization_expression

    @property
    def loop_expression(self) -> Optional[ExpressionStatement]:
        """
        !!! example
            ```solidity
            i++
            ```
            in the following for loop:
            ```solidity
            for (uint i = 0; i < 10; i++) {}
            ```
        Returns:
            Loop expression of the for loop, if any.
        """
        return self._loop_expression

    @property
    @lru_cache(maxsize=2048)
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        ret = set()
        if self.initialization_expression is not None:
            ret |= self.initialization_expression.modifies_state
        if self.condition is not None:
            ret |= self.condition.modifies_state
        if self.loop_expression is not None:
            ret |= self.loop_expression.modifies_state
        return ret

    def statements_iter(self) -> Iterator["StatementAbc"]:
        yield self
        yield from self._body.statements_iter()
        if self._initialization_expression is not None:
            yield from self._initialization_expression.statements_iter()
        if self._loop_expression is not None:
            yield from self._loop_expression.statements_iter()



import logging

import graphviz as gv

from woke.lsp.common_structures import DocumentUri
from woke.lsp.context import LspContext
from woke.lsp.exceptions import LspError
from woke.lsp.protocol_structures import ErrorCodes
from woke.lsp.utils import uri_to_path

logger = logging.getLogger(__name__)


async def generate_linearized_inheritance_graph_handler(
    context: LspContext, uri: DocumentUri, canonical_name: str
) -> str:
    logger.debug(
        f"Linearized inheritance graph for contract {canonical_name} in file {uri} requested"
    )
    await context.compiler.output_ready.wait()

    path = uri_to_path(uri).resolve()

    if path not in context.compiler.source_units:
        raise LspError(ErrorCodes.InternalError, "File not found in compiler output")

    source_unit = context.compiler.source_units[path]
    target_contract = None
    for contract in source_unit.contracts:
        if contract.canonical_name == canonical_name:
            target_contract = contract
            break

    if target_contract is None:
        raise LspError(ErrorCodes.InvalidParams, "Contract not found")

    g = gv.Digraph(f"{canonical_name} linearized inheritance graph")
    g.attr(rankdir=context.config.generator.linearized_inheritance_graph.direction)
    g.attr("node", shape="box")

    prev_node_id = None

    for contract in target_contract.linearized_base_contracts:
        node_id = f"{contract.parent.source_unit_name}_{contract.canonical_name}"
        node_attrs = {}
        if contract.file == path and contract.canonical_name == canonical_name:
            node_attrs["style"] = "filled"

        if context.config.generator.linearized_inheritance_graph.vscode_urls:
            line, column = context.compiler.get_line_pos_from_byte_offset(
                contract.file, contract.name_location[0]
            )
            line += 1
            column += 1
            node_attrs["URL"] = f"vscode://file/{contract.file}:{line}:{column}"

        g.node(node_id, contract.canonical_name, **node_attrs)
        if prev_node_id is not None:
            g.edge(prev_node_id, node_id)
        prev_node_id = node_id
    return g.source



import os
from pathlib import Path
from urllib.parse import quote, unquote, urlparse
from urllib.request import pathname2url, url2pathname


def uri_to_path(uri: str) -> Path:
    path = urlparse(unquote(uri)).path
    path = url2pathname(path)
    return Path(path)


def path_to_uri(path: Path) -> str:
    if os.name == "nt":
        return "file:" + pathname2url(str(path.resolve()))
    else:
        return "file://" + pathname2url(str(path.resolve()))



from __future__ import annotations

from typing import Iterator, Union

from ...nodes import YulForLoop, YulFunctionCall, YulIdentifier, YulLiteral
from ..utils import IrInitTuple
from .abc import YulAbc, YulStatementAbc
from .block import Block
from .function_call import FunctionCall
from .identifier import Identifier
from .literal import Literal


class ForLoop(YulStatementAbc):
    """
    TBD
    """

    _parent: Block
    _body: Block
    _condition: Union[FunctionCall, Identifier, Literal]
    _post: Block
    _pre: Block

    def __init__(self, init: IrInitTuple, for_loop: YulForLoop, parent: YulAbc):
        super().__init__(init, for_loop, parent)
        self._body = Block(init, for_loop.body, self)
        if isinstance(for_loop.condition, YulFunctionCall):
            self._condition = FunctionCall(init, for_loop.condition, self)
        elif isinstance(for_loop.condition, YulIdentifier):
            self._condition = Identifier(init, for_loop.condition, self)
        elif isinstance(for_loop.condition, YulLiteral):
            self._condition = Literal(init, for_loop.condition, self)
        else:
            assert False, f"Unexpected type: {type(for_loop.condition)}"
        self._post = Block(init, for_loop.post, self)
        self._pre = Block(init, for_loop.pre, self)

    def __iter__(self) -> Iterator[YulAbc]:
        yield self
        yield from self._pre
        yield from self._condition
        yield from self._body
        yield from self._post

    @property
    def parent(self) -> Block:
        return self._parent

    @property
    def body(self) -> Block:
        return self._body

    @property
    def condition(self) -> Union[FunctionCall, Identifier, Literal]:
        return self._condition

    @property
    def post(self) -> Block:
        return self._post

    @property
    def pre(self) -> Block:
        return self._pre



from __future__ import annotations

import asyncio
import difflib
import logging
import multiprocessing
import multiprocessing.connection
import platform
import re
import threading
import time
import traceback
from collections import deque
from copy import deepcopy
from functools import lru_cache
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Deque,
    Dict,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Union,
)

from woke.compiler.exceptions import CompilationError

from ..analysis.detectors import detect
from ..core.solidity_version import SolidityVersionRange, SolidityVersionRanges
from ..utils import StrEnum
from ..utils.file_utils import is_relative_to

if TYPE_CHECKING:
    from .server import LspServer

import networkx as nx
from intervaltree import IntervalTree

from woke.ast.ir.meta.source_unit import SourceUnit
from woke.ast.ir.reference_resolver import CallbackParams, ReferenceResolver
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import AstSolc
from woke.compiler import SolcOutput, SolcOutputSelectionEnum
from woke.compiler.compilation_unit import CompilationUnit
from woke.compiler.compiler import SolidityCompiler
from woke.compiler.solc_frontend import SolcOutputError, SolcOutputErrorSeverityEnum
from woke.config import WokeConfig
from woke.lsp.document_sync import (
    DidChangeTextDocumentParams,
    DidCloseTextDocumentParams,
    DidOpenTextDocumentParams,
)
from woke.lsp.utils.uri import path_to_uri, uri_to_path

from ..svm import SolcVersionManager
from .common_structures import (
    CreateFilesParams,
    DeleteFile,
    DeleteFilesParams,
    Diagnostic,
    DiagnosticRelatedInformation,
    DiagnosticSeverity,
    DocumentUri,
    Location,
    MessageType,
    Position,
    Range,
    RenameFilesParams,
)

logger = logging.getLogger(__name__)


ENCODING = "utf-16-le"


def _binary_search(lines: List[Tuple[bytes, int]], x: int) -> int:
    l = 0
    r = len(lines)

    while l < r:
        mid = l + (r - l) // 2
        if lines[mid][1] < x + 1:
            l = mid + 1
        else:
            r = mid

    return l - 1


def _out_edge_bfs(cu: CompilationUnit, start: Iterable[Path], out: Set[Path]) -> None:
    processed: Set[str] = set()
    for path in start:
        processed.update(cu.path_to_source_unit_names(path))
    out.update(start)

    queue: Deque[str] = deque(processed)
    while len(queue):
        node = queue.pop()
        for out_edge in cu.graph.out_edges(
            node  # pyright: ignore reportGeneralTypeIssues
        ):
            from_, to = out_edge  # pyright: ignore reportGeneralTypeIssues
            if to not in processed:
                processed.add(to)
                queue.append(to)
                out.add(cu.source_unit_name_to_path(to))


class VersionedFile(NamedTuple):
    text: str
    version: Optional[int]


class CustomFileChangeCommand(StrEnum):
    FORCE_RECOMPILE = "force_recompile"
    FORCE_RERUN_DETECTORS = "force_rerun_detectors"


class LspCompiler:
    __config: WokeConfig
    __svm: SolcVersionManager
    __server: LspServer
    __file_changes_queue: asyncio.Queue
    __diagnostic_queue: asyncio.Queue
    __discovered_files: Set[Path]
    __deleted_files: Set[Path]
    __opened_files: Dict[Path, VersionedFile]
    __modified_files: Set[Path]
    __force_compile_files: Set[Path]
    __compiler: SolidityCompiler
    __output_contents: Dict[Path, VersionedFile]
    __compilation_errors: Dict[Path, Set[Diagnostic]]
    __last_successful_compilation_contents: Dict[Path, VersionedFile]
    __interval_trees: Dict[Path, IntervalTree]
    __source_units: Dict[Path, SourceUnit]
    __last_compilation_interval_trees: Dict[Path, IntervalTree]
    __last_compilation_source_units: Dict[Path, SourceUnit]
    __last_graph: nx.DiGraph
    __line_indexes: Dict[Path, List[Tuple[bytes, int]]]
    __perform_files_discovery: bool
    __force_run_detectors: bool

    __ir_reference_resolver: ReferenceResolver

    __output_ready: asyncio.Event

    def __init__(
        self,
        server: LspServer,
        diagnostic_queue: asyncio.Queue,
        perform_files_discovery: bool,
    ):
        self.__server = server
        self.__file_changes_queue = asyncio.Queue()
        self.__diagnostic_queue = diagnostic_queue
        self.__stop_event = threading.Event()
        self.__discovered_files = set()
        self.__deleted_files = set()
        self.__opened_files = {}
        self.__modified_files = set()
        self.__force_compile_files = set()
        self.__interval_trees = {}
        self.__source_units = {}
        self.__last_compilation_interval_trees = {}
        self.__last_compilation_source_units = {}
        self.__last_graph = nx.DiGraph()
        self.__line_indexes = {}
        self.__output_contents = dict()
        self.__compilation_errors = dict()
        self.__last_successful_compilation_contents = dict()
        self.__output_ready = asyncio.Event()
        self.__perform_files_discovery = perform_files_discovery
        self.__force_run_detectors = False

        self.__ir_reference_resolver = ReferenceResolver()

    async def run(self, config: WokeConfig):
        self.__config = config
        self.__svm = SolcVersionManager(config)
        self.__compiler = SolidityCompiler(config)
        await self.__compilation_loop()

    @property
    def output_ready(self) -> asyncio.Event:
        return self.__output_ready

    @property
    def ir_reference_resolver(self) -> ReferenceResolver:
        return self.__ir_reference_resolver

    @property
    def interval_trees(self) -> Dict[Path, IntervalTree]:
        return self.__interval_trees

    @property
    def source_units(self) -> Dict[Path, SourceUnit]:
        return self.__source_units

    @property
    def last_compilation_interval_trees(self) -> Dict[Path, IntervalTree]:
        return self.__last_compilation_interval_trees

    @property
    def last_compilation_source_units(self) -> Dict[Path, SourceUnit]:
        return self.__last_compilation_source_units

    @property
    def last_graph(self) -> nx.DiGraph:
        return self.__last_graph

    @lru_cache(maxsize=128)
    def _compute_diff_interval_tree(
        self, a: VersionedFile, b: VersionedFile
    ) -> IntervalTree:
        seq_matcher = difflib.SequenceMatcher(None, a.text, b.text)
        interval_tree = IntervalTree()
        for tag, i1, i2, j1, j2 in seq_matcher.get_opcodes():
            if tag == "equal":
                continue
            interval_tree.addi(i1, i2 + 1, (tag, j1, j2 + 1))
        return interval_tree

    def get_last_compilation_forward_changes(
        self, path: Path
    ) -> Optional[IntervalTree]:
        if path not in self.__last_successful_compilation_contents:
            return None
        return self._compute_diff_interval_tree(
            self.__last_successful_compilation_contents[path],
            self.get_compiled_file(path),
        )

    def get_last_compilation_backward_changes(
        self, path: Path
    ) -> Optional[IntervalTree]:
        if path not in self.__last_successful_compilation_contents:
            return None
        return self._compute_diff_interval_tree(
            self.get_compiled_file(path),
            self.__last_successful_compilation_contents[path],
        )

    async def add_change(
        self,
        change: Union[
            DidOpenTextDocumentParams,
            DidChangeTextDocumentParams,
            DidCloseTextDocumentParams,
            CreateFilesParams,
            RenameFilesParams,
            DeleteFilesParams,
        ],
    ) -> None:
        self.output_ready.clear()
        await self.__file_changes_queue.put(change)

    async def force_recompile(self) -> None:
        self.__output_ready.clear()
        await self.__file_changes_queue.put(CustomFileChangeCommand.FORCE_RECOMPILE)

    async def force_rerun_detectors(self) -> None:
        await self.__file_changes_queue.put(
            CustomFileChangeCommand.FORCE_RERUN_DETECTORS
        )

    def get_compiled_file(self, file: Union[Path, str]) -> VersionedFile:
        if isinstance(file, str):
            file = uri_to_path(file)
        if file not in self.__output_contents:
            self.__output_contents[file] = VersionedFile(
                file.read_bytes().decode(encoding="utf-8"), None
            )
        return self.__output_contents[file]

    def get_line_pos_from_byte_offset(
        self, file: Path, byte_offset: int
    ) -> Tuple[int, int]:
        if file not in self.__line_indexes:
            self.__setup_line_index(file)

        encoded_lines = self.__line_indexes[file]
        line_num = _binary_search(encoded_lines, byte_offset)
        line_data, prefix_sum = encoded_lines[line_num]
        line_offset = byte_offset - prefix_sum
        return line_num, line_offset

    def get_range_from_byte_offsets(
        self, file: Path, byte_offsets: Tuple[int, int]
    ) -> Range:
        start_line, start_column = self.get_line_pos_from_byte_offset(
            file, byte_offsets[0]
        )
        end_line, end_column = self.get_line_pos_from_byte_offset(file, byte_offsets[1])

        return Range(
            start=Position(line=start_line, character=start_column),
            end=Position(line=end_line, character=end_column),
        )

    def get_byte_offset_from_line_pos(self, file: Path, line: int, col: int) -> int:
        if file not in self.__line_indexes:
            self.__setup_line_index(file)

        encoded_lines = self.__line_indexes[file]
        line_bytes, prefix = encoded_lines[line]
        line_offset = len(line_bytes.decode("utf-8")[:col].encode("utf-8"))
        return prefix + line_offset

    async def _handle_change(
        self,
        change: Union[
            DidOpenTextDocumentParams,
            DidCloseTextDocumentParams,
            DidChangeTextDocumentParams,
            None,
        ],
    ) -> None:
        if isinstance(change, CustomFileChangeCommand):
            if change == CustomFileChangeCommand.FORCE_RECOMPILE:
                for file in self.__discovered_files:
                    # clear diagnostics
                    await self.__diagnostic_queue.put((file, set()))

                self.__discovered_files.clear()
                self.__compilation_errors.clear()
                if self.__perform_files_discovery:
                    for file in self.__config.project_root_path.rglob("**/*.sol"):
                        if not self.__file_ignored(file) and file.is_file():
                            self.__discovered_files.add(file.resolve())

                for file in self.__opened_files.keys():
                    if not self.__file_ignored(file):
                        self.__discovered_files.add(file)

                self.__force_compile_files.update(self.__discovered_files)
            elif change == CustomFileChangeCommand.FORCE_RERUN_DETECTORS:
                self.__force_run_detectors = True
        elif isinstance(change, CreateFilesParams):
            for file in change.files:
                path = uri_to_path(file.uri)
                if (
                    path not in self.__discovered_files
                    and not self.__file_ignored(path)
                    and path.is_file()
                    and path.suffix == ".sol"
                ):
                    self.__discovered_files.add(path)
                    self.__force_compile_files.add(path)
        elif isinstance(change, RenameFilesParams):
            for rename in change.files:
                old_path = uri_to_path(rename.old_uri)
                self.__deleted_files.add(old_path)
                self.__discovered_files.discard(old_path)
                self.__opened_files.pop(old_path, None)

                new_path = uri_to_path(rename.new_uri)
                if (
                    new_path not in self.__discovered_files
                    and not self.__file_ignored(new_path)
                    and new_path.is_file()
                    and new_path.suffix == ".sol"
                ):
                    self.__discovered_files.add(new_path)
                    self.__force_compile_files.add(new_path)
        elif isinstance(change, DeleteFilesParams):
            for delete in change.files:
                path = uri_to_path(delete.uri)
                self.__deleted_files.add(path)
                self.__discovered_files.discard(path)
                self.__opened_files.pop(path, None)
        elif isinstance(change, DidOpenTextDocumentParams):
            path = uri_to_path(change.text_document.uri).resolve()
            self.__opened_files[path] = VersionedFile(
                change.text_document.text, change.text_document.version
            )
            if (
                path not in self.__discovered_files
                and not self.__file_ignored(path)
                and path.is_file()
                and path.suffix == ".sol"
            ):
                self.__discovered_files.add(path)
                self.__force_compile_files.add(path)
            elif change.text_document.text != self.get_compiled_file(path).text:
                self.__force_compile_files.add(path)
            else:
                self.__output_contents[path] = self.__opened_files[path]

        elif isinstance(change, DidCloseTextDocumentParams):
            pass
        elif isinstance(change, DidChangeTextDocumentParams):
            path = uri_to_path(change.text_document.uri).resolve()
            self.__modified_files.add(path)

            for content_change in change.content_changes:
                start = content_change.range.start
                end = content_change.range.end

                # str.splitlines() removes empty lines => cannot be used
                # str.split() removes separators => cannot be used
                tmp_lines = re.split(r"(\r?\n)", self.__opened_files[path].text)
                tmp_lines2: List[str] = []
                for line in tmp_lines:
                    if line in {"\r\n", "\n"}:
                        tmp_lines2[-1] += line
                    else:
                        tmp_lines2.append(line)

                lines: List[bytearray] = [
                    bytearray(line.encode(ENCODING)) for line in tmp_lines2
                ]

                if start.line == end.line:
                    line = lines[start.line]
                    line[start.character * 2 : end.character * 2] = b""
                    line[
                        start.character * 2 : start.character * 2
                    ] = content_change.text.encode(ENCODING)
                else:
                    start_line = lines[start.line]
                    end_line = lines[end.line]
                    start_line[start.character * 2 :] = content_change.text.encode(
                        ENCODING
                    )
                    end_line[: end.character * 2] = b""

                    for i in range(start.line + 1, end.line):
                        lines[i] = bytearray(b"")

                self.__opened_files[path] = VersionedFile(
                    "".join(line.decode(ENCODING) for line in lines),
                    change.text_document.version,
                )
        else:
            raise Exception("Unknown change type")

    async def __compile(
        self,
        files_to_compile: AbstractSet[Path],
        full_compile: bool = True,
    ) -> None:
        target_version = self.__config.compiler.solc.target_version
        min_version = self.__config.min_solidity_version
        max_version = self.__config.max_solidity_version
        if target_version is not None and target_version < min_version:
            await self.__server.log_message(
                f"The minimum supported version of Solidity is {min_version}. Version {target_version} is selected in settings.",
                MessageType.ERROR,
            )
            for file in self.__discovered_files:
                # clear diagnostics
                await self.__diagnostic_queue.put((file, set()))
            self.__interval_trees.clear()
            self.__source_units.clear()
            return
        if target_version is not None and target_version > max_version:
            await self.__server.log_message(
                f"The maximum supported version of Solidity is {max_version}. Version {target_version} is selected in settings.",
                MessageType.ERROR,
            )
            for file in self.__discovered_files:
                # clear diagnostics
                await self.__diagnostic_queue.put((file, set()))
            self.__interval_trees.clear()
            self.__source_units.clear()
            return

        try:
            if full_compile:
                graph, _ = self.__compiler.build_graph(
                    self.__discovered_files,
                    {path: info.text for path, info in self.__opened_files.items()},
                    True,
                )
                self.__last_graph = graph
            else:
                graph, _ = self.__compiler.build_graph(
                    files_to_compile,
                    {path: info.text for path, info in self.__opened_files.items()},
                    True,
                )
        except CompilationError as e:
            await self.__server.log_message(str(e), MessageType.ERROR)
            for file in self.__discovered_files:
                # clear diagnostics
                await self.__diagnostic_queue.put((file, set()))
            self.__interval_trees.clear()
            self.__source_units.clear()
            return

        compilation_units = self.__compiler.build_compilation_units_maximize(graph)

        # filter out only compilation units that need to be compiled
        compilation_units = [
            cu
            for cu in compilation_units
            if (cu.files & files_to_compile)
            or cu.contains_unresolved_file(self.__deleted_files, self.__config)
        ]
        if len(compilation_units) == 0:
            return

        build_settings = self.__compiler.create_build_settings(
            [SolcOutputSelectionEnum.AST]
        )

        # optimization - merge compilation units that can be compiled together
        if all(len(cu.versions) for cu in compilation_units):
            compilation_units = sorted(
                compilation_units, key=lambda cu: cu.versions.version_ranges[0].lower
            )

            merged_compilation_units: List[CompilationUnit] = []
            source_unit_names: Set = set()
            versions = SolidityVersionRanges(
                [SolidityVersionRange(None, None, None, None)]
            )

            for cu in compilation_units:
                if versions & cu.versions:
                    source_unit_names |= cu.source_unit_names
                    versions &= cu.versions
                else:
                    merged_compilation_units.append(
                        CompilationUnit(
                            graph.subgraph(source_unit_names).copy(),
                            versions,
                        )
                    )
                    source_unit_names = set(cu.source_unit_names)
                    versions = cu.versions

            merged_compilation_units.append(
                CompilationUnit(
                    graph.subgraph(source_unit_names).copy(),
                    versions,
                )
            )

            compilation_units = merged_compilation_units

        target_versions = []
        skipped_compilation_units = []
        for compilation_unit in compilation_units:
            target_version = self.__config.compiler.solc.target_version
            if target_version is not None:
                if target_version not in compilation_unit.versions:
                    await self.__server.log_message(
                        f"Unable to compile the following files with solc version `{target_version}` set in config:\n"
                        + "\n".join(
                            path_to_uri(path) for path in compilation_unit.files
                        ),
                        MessageType.ERROR,
                    )
                    skipped_compilation_units.append(compilation_unit)
                    continue
            else:
                # use the latest matching version
                matching_versions = [
                    version
                    for version in reversed(self.__svm.list_all())
                    if version in compilation_unit.versions
                ]
                if len(matching_versions) == 0:
                    await self.__server.log_message(
                        f"Unable to find a matching version of Solidity for the following files:\n"
                        + "\n".join(
                            path_to_uri(path) for path in compilation_unit.files
                        ),
                        MessageType.ERROR,
                    )
                    skipped_compilation_units.append(compilation_unit)
                    continue
                try:
                    target_version = next(
                        version
                        for version in matching_versions
                        if version <= max_version
                    )
                except StopIteration:
                    await self.__server.log_message(
                        f"The maximum supported version of Solidity is {max_version}, unable to compile the following files:\n"
                        + "\n".join(
                            path_to_uri(path) for path in compilation_unit.files
                        ),
                        MessageType.ERROR,
                    )
                    skipped_compilation_units.append(compilation_unit)
                    continue

                if target_version < min_version:
                    await self.__server.log_message(
                        f"The minimum supported version of Solidity is {min_version}, unable to compile the following files:\n"
                        + "\n".join(
                            path_to_uri(path) for path in compilation_unit.files
                        ),
                        MessageType.ERROR,
                    )
                    skipped_compilation_units.append(compilation_unit)
                    continue
            target_versions.append(target_version)

        for version in set(target_versions):
            if not self.__svm.installed(version):
                progress_token = await self.__server.progress_begin(
                    "Downloading", f"solc {version}", 0
                )
                if progress_token is not None:

                    async def on_progress(downloaded: int, total: int) -> None:
                        assert progress_token is not None
                        await self.__server.progress_report(
                            progress_token,
                            f"solc {version}",
                            (100 * downloaded) // total,
                        )

                    await self.__svm.install(version, progress=on_progress)
                    await self.__server.progress_end(progress_token)
                else:
                    await self.__svm.install(version)

        for compilation_unit in skipped_compilation_units:
            for file in compilation_unit.files:
                # clear diagnostics
                await self.__diagnostic_queue.put((file, set()))
                if file in self.__interval_trees:
                    self.__interval_trees.pop(file)
                if file in self.__source_units:
                    self.__ir_reference_resolver.run_destroy_callbacks(file)
                    self.__source_units.pop(file)
            compilation_units.remove(compilation_unit)

        progress_token = await self.__server.progress_begin(
            "Compiling", f"0/{len(compilation_units)}", 0
        )

        tasks = []
        for compilation_unit, target_version in zip(compilation_units, target_versions):
            task = self.__server.create_task(
                self.__compiler.compile_unit_raw(
                    compilation_unit,
                    target_version,
                    build_settings,
                )
            )
            tasks.append(task)

        # wait for compilation of all compilation units
        try:
            ret = await asyncio.gather(*tasks)
        except Exception as e:
            for task in tasks:
                task.cancel()
            await self.__server.log_message(str(e), MessageType.ERROR)
            if progress_token is not None:
                await self.__server.progress_end(progress_token)

            for file in self.__discovered_files:
                # clear diagnostics
                await self.__diagnostic_queue.put((file, set()))
            self.__interval_trees.clear()
            self.__source_units.clear()
            return

        errors_without_location: Set[SolcOutputError] = set()
        errors_per_file: Dict[Path, Set[Diagnostic]] = deepcopy(
            self.__compilation_errors
        )

        for cu, solc_output in zip(compilation_units, ret):
            for file in cu.files:
                errors_per_file[file] = set()
            for error in solc_output.errors:
                if error.source_location is None:
                    errors_without_location.add(error)

        for error in errors_without_location:
            if error.severity == SolcOutputErrorSeverityEnum.ERROR:
                error_type = MessageType.ERROR
            elif error.severity == SolcOutputErrorSeverityEnum.WARNING:
                error_type = MessageType.WARNING
            elif error.severity == SolcOutputErrorSeverityEnum.INFO:
                error_type = MessageType.INFO
            else:
                error_type = MessageType.LOG
            await self.__server.show_message(error.message, error_type)
            await self.__server.log_message(error.message, error_type)

        if (
            any(
                e.severity == SolcOutputErrorSeverityEnum.ERROR
                for e in errors_without_location
            )
            > 0
        ):
            if progress_token is not None:
                await self.__server.progress_end(progress_token)

            for file in self.__discovered_files:
                # clear diagnostics
                await self.__diagnostic_queue.put((file, set()))
            self.__interval_trees.clear()
            self.__source_units.clear()
            return

        # files passed as files_to_compile and files importing them
        files_to_recompile = set(
            graph.nodes[n[1]]["path"]  # pyright: ignore reportGeneralTypeIssues
            for n in nx.edge_bfs(
                graph,
                [
                    source_unit_name
                    for source_unit_name in graph.nodes  # pyright: ignore reportGeneralTypeIssues
                    if graph.nodes[  # pyright: ignore reportGeneralTypeIssues
                        source_unit_name
                    ][  # pyright: ignore reportGeneralTypeIssues
                        "path"
                    ]
                    in files_to_compile  # pyright: ignore reportGeneralTypeIssues
                ],
            )
        ) | set(files_to_compile)

        for deleted_file in self.__deleted_files:
            await self.__diagnostic_queue.put((deleted_file, []))
            if deleted_file in self.__source_units:
                self.__ir_reference_resolver.run_destroy_callbacks(deleted_file)
                self.__source_units.pop(deleted_file)

        successful_compilation_units = []
        for cu, solc_output in zip(compilation_units, ret):
            for file in cu.files:
                if file in self.__line_indexes:
                    self.__line_indexes.pop(file)
                if file in self.__opened_files:
                    self.__output_contents[file] = self.__opened_files[file]

            errored_files: Set[Path] = set()

            for error in solc_output.errors:
                if error.source_location is not None:
                    path = cu.source_unit_name_to_path(error.source_location.file)
                    errors_per_file[path].add(
                        self.__solc_error_to_diagnostic(error, path, cu)
                    )
                    if error.severity == SolcOutputErrorSeverityEnum.ERROR:
                        errored_files.add(path)
                else:
                    # whole compilation unit errored
                    if error.severity == SolcOutputErrorSeverityEnum.ERROR:
                        errored_files.update(cu.files)

            _out_edge_bfs(cu, errored_files, errored_files)

            for file in errored_files:
                files_to_recompile.discard(file)
                # an error occurred during compilation
                # AST still may be provided, but it must NOT be parsed (pydantic model is not defined for this case)
                if file in self.__source_units:
                    self.__ir_reference_resolver.run_destroy_callbacks(file)
                    self.__source_units.pop(file)
                if file in self.__interval_trees:
                    self.__interval_trees.pop(file)

            if len(errored_files) == 0:
                successful_compilation_units.append((cu, solc_output))

        # destroy callbacks for IR nodes that will be replaced by new ones must be executed before
        # new IR nodes are created
        callback_processed_files: Set[Path] = set()
        for cu, solc_output in successful_compilation_units:
            # files requested to be compiled and files that import these files (even indirectly)
            recompiled_files: Set[Path] = set()
            _out_edge_bfs(cu, files_to_compile & cu.files, recompiled_files)

            # run destroy callbacks for files that were recompiled and their IR nodes will be replaced
            for source_unit_name in solc_output.sources.keys():
                path = cu.source_unit_name_to_path(source_unit_name)
                if (
                    path in self.__source_units and path not in recompiled_files
                ) or path in callback_processed_files:
                    continue
                callback_processed_files.add(path)
                self.__ir_reference_resolver.run_destroy_callbacks(path)

        # clear indexed node types responsible for handling multiple structurally different ASTs for the same file
        self.__ir_reference_resolver.clear_indexed_nodes(files_to_recompile)

        processed_files: Set[Path] = set()
        for cu_index, (cu, solc_output) in enumerate(successful_compilation_units):
            # files requested to be compiled and files that import these files (even indirectly)
            recompiled_files: Set[Path] = set()
            _out_edge_bfs(cu, files_to_compile & cu.files, recompiled_files)

            for source_unit_name, raw_ast in solc_output.sources.items():
                path: Path = cu.source_unit_name_to_path(source_unit_name)
                ast = AstSolc.parse_obj(raw_ast.ast)

                self.__ir_reference_resolver.index_nodes(ast, path, cu.hash)

                files_to_recompile.discard(path)

                # give a chance to other tasks (LSP requests) to be processed
                await asyncio.sleep(0)

                if (
                    path in self.__source_units and path not in recompiled_files
                ) or path in processed_files:
                    continue
                processed_files.add(path)

                interval_tree = IntervalTree()
                init = IrInitTuple(
                    path,
                    self.get_compiled_file(path).text.encode("utf-8"),
                    cu,
                    interval_tree,
                    self.__ir_reference_resolver,
                    None,
                )
                self.__source_units[path] = SourceUnit(init, ast)
                self.__interval_trees[path] = interval_tree

                self.__last_compilation_source_units[path] = self.__source_units[path]
                self.__last_compilation_interval_trees[path] = self.__interval_trees[
                    path
                ]

                if path in self.__opened_files:
                    self.__last_successful_compilation_contents[
                        path
                    ] = self.__opened_files[path]
                else:
                    self.__last_successful_compilation_contents[path] = VersionedFile(
                        cu.graph.nodes[  # pyright: ignore reportGeneralTypeIssues
                            source_unit_name
                        ]["content"],
                        None,
                    )

            self.__ir_reference_resolver.run_post_process_callbacks(
                CallbackParams(
                    interval_trees=self.__interval_trees,
                    source_units=self.__source_units,
                )
            )

            if progress_token is not None:
                await self.__server.progress_report(
                    progress_token,
                    f"{cu_index + 1}/{len(compilation_units)}",
                    ((cu_index + 1) * 100) // len(compilation_units),
                )

        if progress_token is not None:
            await self.__server.progress_end(progress_token)

        self.__compilation_errors = deepcopy(errors_per_file)

        # send compiler warnings and errors first without waiting for detectors to finish
        for path, errors in errors_per_file.items():
            await self.__diagnostic_queue.put((path, errors))

        await self.__run_detectors_wrapper()

        if len(files_to_recompile) > 0:
            # avoid infinite recursion
            if files_to_recompile != files_to_compile or full_compile:
                await self.__compile(files_to_recompile, False)

    async def __run_detectors_wrapper(self) -> None:
        if platform.system() == "Linux":
            # run detectors in a separate process
            parent_conn, child_conn = multiprocessing.Pipe()
            p = multiprocessing.Process(
                target=self.__run_detectors,
                args=(child_conn,),
            )
            p.start()
            while True:
                if parent_conn.poll():
                    errors_per_file = parent_conn.recv()
                    # send both compiler and detector warnings and errors
                    for path, errors in errors_per_file.items():
                        await self.__diagnostic_queue.put((path, errors))
                    break
                elif not p.is_alive():
                    break

                await asyncio.sleep(0.1)
        else:
            # on Windows, forking is not available,
            # on macOS, forking is available, but it's not recommended
            # pickling and unpickling of whole IR model with spawn/forkserver would be too slow

            # not in a subprocess, use try/except to avoid crashing the server by detectors
            try:
                errors_per_file = self.__run_detectors(None)
                # send both compiler and detector warnings and errors
                for path, errors in errors_per_file.items():
                    await self.__diagnostic_queue.put((path, errors))
            except Exception:
                await self.__server.log_message(
                    f"Exception occurred during running detectors:\n{traceback.format_exc()}",
                    MessageType.ERROR,
                )

    def __run_detectors(
        self, conn: Optional[multiprocessing.connection.Connection]
    ) -> Dict[Path, Set[Diagnostic]]:
        errors_per_file: Dict[Path, Set[Diagnostic]]
        if conn is not None:
            # in forked process, original dictionary won't be modified anyway
            errors_per_file = self.__compilation_errors
        else:
            errors_per_file = deepcopy(self.__compilation_errors)

        if self.__config.lsp.detectors.enable:
            for detection in detect(self.__config, self.__source_units):
                file = detection.result.ir_node.file
                if len(detection.result.related_info) > 0:
                    related_info = [
                        DiagnosticRelatedInformation(
                            location=Location(
                                uri=DocumentUri(path_to_uri(info.ir_node.file)),
                                range=self.get_range_from_byte_offsets(
                                    info.ir_node.file,
                                    info.lsp_range
                                    if info.lsp_range is not None
                                    else info.ir_node.byte_location,
                                ),
                            ),
                            message=info.message,
                        )
                        for info in detection.result.related_info
                    ]
                else:
                    related_info = None

                if file not in errors_per_file:
                    errors_per_file[file] = set()
                errors_per_file[file].add(
                    Diagnostic(
                        range=self.get_range_from_byte_offsets(
                            file,
                            detection.result.lsp_range
                            if detection.result.lsp_range is not None
                            else detection.result.ir_node.byte_location,
                        ),
                        severity=DiagnosticSeverity.WARNING,
                        source="Woke",
                        message=detection.result.message,
                        code=detection.code,
                        related_information=related_info,
                    )
                )

        if conn is not None:
            conn.send(errors_per_file)
        return errors_per_file

    async def __compilation_loop(self):
        if self.__perform_files_discovery:
            # perform Solidity files discovery
            for file in self.__config.project_root_path.rglob("**/*.sol"):
                if not self.__file_ignored(file) and file.is_file():
                    self.__discovered_files.add(file.resolve())

            # perform initial compilation
            await self.__compile(self.__discovered_files)

        if self.__file_changes_queue.empty():
            self.output_ready.set()

        while True:
            change = await self.__file_changes_queue.get()
            start = time.perf_counter()
            await self._handle_change(change)
            while True:
                try:
                    change = self.__file_changes_queue.get_nowait()
                    start = time.perf_counter()
                    await self._handle_change(change)
                except asyncio.QueueEmpty:
                    if (
                        time.perf_counter() - start
                        > self.__config.lsp.compilation_delay
                    ):
                        break
                    await asyncio.sleep(0.1)

            # run the compilation
            if (
                len(self.__force_compile_files) > 0
                or len(self.__modified_files) > 0
                or len(self.__deleted_files) > 0
            ):
                await self.__compile(
                    self.__force_compile_files.union(self.__modified_files)
                )

                self.__force_compile_files.clear()
                self.__modified_files.clear()
                self.__deleted_files.clear()

            if self.__force_run_detectors:
                await self.__run_detectors_wrapper()
                self.__force_run_detectors = False

            if self.__file_changes_queue.empty():
                self.output_ready.set()

    def __setup_line_index(self, file: Path):
        content = self.get_compiled_file(file).text
        tmp_lines = re.split(r"(\r?\n)", content)
        lines: List[str] = []
        for line in tmp_lines:
            if line in {"\r\n", "\n"}:
                lines[-1] += line
            else:
                lines.append(line)

        # UTF-8 encoded lines with prefix length
        encoded_lines: List[Tuple[bytes, int]] = []
        prefix_sum = 0
        for line in lines:
            encoded_line = line.encode("utf-8")
            encoded_lines.append((encoded_line, prefix_sum))
            prefix_sum += len(encoded_line)
        self.__line_indexes[file] = encoded_lines

    def __solc_error_to_diagnostic(
        self, error: SolcOutputError, path: Path, cu: CompilationUnit
    ) -> Diagnostic:
        assert error.source_location is not None
        if error.severity == SolcOutputErrorSeverityEnum.ERROR:
            severity = DiagnosticSeverity.ERROR
        elif error.severity == SolcOutputErrorSeverityEnum.WARNING:
            severity = DiagnosticSeverity.WARNING
        else:
            severity = DiagnosticSeverity.INFORMATION

        if error.source_location.start >= 0 and error.source_location.end >= 0:
            range_ = self.get_range_from_byte_offsets(
                path, (error.source_location.start, error.source_location.end)
            )
        else:
            range_ = Range(
                start=Position(line=0, character=0),
                end=Position(line=0, character=0),
            )

        diag = Diagnostic(
            range=range_,
            severity=severity,
            code=error.error_code,
            source="Woke(solc)",
            message=error.message,
        )
        if (
            error.secondary_source_locations is not None
            and len(error.secondary_source_locations) > 0
        ):
            related_info = []
            for secondary_source_location in error.secondary_source_locations:
                if (
                    secondary_source_location.file is None
                    or secondary_source_location.start is None
                    or secondary_source_location.end is None
                ):
                    continue

                if (
                    secondary_source_location.start >= 0
                    and secondary_source_location.end >= 0
                ):
                    range_ = self.get_range_from_byte_offsets(
                        cu.source_unit_name_to_path(secondary_source_location.file),
                        (
                            secondary_source_location.start,
                            secondary_source_location.end,
                        ),
                    )
                else:
                    range_ = Range(
                        start=Position(line=0, character=0),
                        end=Position(line=0, character=0),
                    )

                related_info.append(
                    DiagnosticRelatedInformation(
                        location=Location(
                            uri=DocumentUri(
                                path_to_uri(
                                    cu.source_unit_name_to_path(
                                        secondary_source_location.file
                                    )
                                )
                            ),
                            range=range_,
                        ),
                        message=secondary_source_location.message,
                    )
                )
            diag.related_information = related_info
        return diag

    def __file_ignored(self, path: Path) -> bool:
        return any(
            is_relative_to(path, p) for p in self.__config.compiler.solc.ignore_paths
        )



from typing import List, Set

from woke.analysis.detectors import DetectorAbc, DetectorResult, detector
from woke.analysis.detectors.utils import check_all_return_params_set
from woke.ast.enums import GlobalSymbolsEnum
from woke.ast.ir.declaration.function_definition import FunctionDefinition
from woke.ast.ir.expression.function_call import FunctionCall
from woke.ast.ir.statement.expression_statement import ExpressionStatement
from woke.ast.ir.statement.inline_assembly import InlineAssembly
from woke.ast.ir.statement.return_statement import Return
from woke.ast.ir.statement.revert_statement import RevertStatement
from woke.ast.ir.yul.expression_statement import (
    ExpressionStatement as YulExpressionStatement,
)
from woke.ast.ir.yul.function_call import FunctionCall as YulFunctionCall


def check_missing_return(node: FunctionDefinition) -> List[DetectorResult]:
    cfg = node.cfg
    assert cfg is not None
    end = cfg.end_block
    graph = cfg.graph
    detections = []

    for block in graph.predecessors(end):
        has_return = False
        for stmt in block.statements:
            if isinstance(stmt, ExpressionStatement):
                if (
                    isinstance(stmt.expression, FunctionCall)
                    and stmt.expression.function_called == GlobalSymbolsEnum.REVERT
                ):
                    has_return = True
                    break
                elif isinstance(stmt.expression, Return) or isinstance(
                    stmt.expression, RevertStatement
                ):
                    has_return = True
                    break
            elif isinstance(stmt, Return) or isinstance(stmt, RevertStatement):
                has_return = True
                break
            elif isinstance(stmt, InlineAssembly) and stmt.yul_block is not None:
                for yul_stmt in stmt.yul_block.statements:
                    if (
                        isinstance(yul_stmt, YulExpressionStatement)
                        and isinstance(yul_stmt.expression, YulFunctionCall)
                        and yul_stmt.expression.function_name.name
                        in ("revert", "return")
                    ):
                        has_return = True
                        break
        if not has_return:
            has_return_params_named = True
            for param in node.return_parameters.parameters:
                if param.name == "":
                    has_return_params_named = False
                    break

            if not has_return_params_named:
                detections.append(
                    DetectorResult(
                        node,
                        "Not all code paths have return or revert statement",
                        lsp_range=node.name_location,
                    )
                )
            else:
                solved, params = check_all_return_params_set(
                    node.return_parameters.parameters, graph, block, cfg.start_block
                )
                if not solved:
                    detections.append(
                        DetectorResult(
                            node,
                            "Not all code paths have return or revert statement and the return values "
                            "are not set either",
                            lsp_range=node.name_location,
                        )
                    )
    return detections


@detector(-1030, "missing-return")
class MissingReturnDetector(DetectorAbc):
    """
    Detector that checks if all possible paths have a return or revert statement or have all
    return values set
    """

    _detections: Set[DetectorResult]

    def __init__(self):
        self._detections = set()

    def report(self) -> List[DetectorResult]:
        return list(self._detections)

    def visit_function_definition(self, node: FunctionDefinition):
        if (
            node.body is None
            or len(node.body.statements) == 0
            or len(node.return_parameters.parameters) == 0
        ):
            return

        for det in check_missing_return(node):
            self._detections.add(det)



from .woke_config import UnsupportedPlatformError, WokeConfig

__doc__ = """This module handles config file management. Each config option has its default value.
There are two main sources of config files:
* `config.toml` global config file in the Woke root directory ($HOME/.config/Woke on macOS and Linux, $HOME/Woke on Windows)
* `woke.toml` project-specific config file present in a project root directory

There may be additional config files included with the `subconfigs` top-level config key. Paths in the `subconfigs` key can
be both relative and absolute.

Config options can be overridden. Imported config options override the options in the original file. Order of files
listed in `subconfigs` also matters. Latter files in the list override earlier files. Config options loaded from the
global `config.toml` file can be overridden by options supplied through project-specific config files.

While this module enforces valid syntax of config files, it does not (and cannot) verify the semantics of the provided
config values. Extra config keys that are not specified in the documentation are forbidden."""



from .assignment import Assignment
from .block import Block
from .break_statement import Break
from .case_statement import Case
from .continue_statement import Continue
from .expression_statement import ExpressionStatement
from .for_loop import ForLoop
from .function_call import FunctionCall
from .function_definition import FunctionDefinition
from .identifier import Identifier
from .if_statement import If
from .leave import Leave
from .literal import Literal
from .switch import Switch
from .typed_name import TypedName
from .variable_declaration import VariableDeclaration



from collections import deque
from itertools import chain
from typing import Deque, List, Set

from woke.analysis.detectors import DetectorAbc, DetectorResult, detector
from woke.ast.enums import GlobalSymbolsEnum, StateMutability
from woke.ast.ir.declaration.function_definition import FunctionDefinition
from woke.ast.ir.expression.function_call import FunctionCall
from woke.ast.ir.expression.member_access import MemberAccess
from woke.ast.ir.meta.identifier_path import IdentifierPathPart
from woke.ast.ir.statement.inline_assembly import ExternalReference


@detector(-1031, "msg-value-nonpayable-function")
class MsgValueNonpayableFunctionDetector(DetectorAbc):
    _detections: List[DetectorResult]

    def __init__(self):
        self._detections = []

    def report(self) -> List[DetectorResult]:
        return self._detections

    def visit_member_access(self, node: MemberAccess):
        if node.referenced_declaration != GlobalSymbolsEnum.MSG_VALUE:
            return

        func = node
        while func is not None:
            if isinstance(func, FunctionDefinition):
                break
            func = func.parent

        if func is None:
            return

        if func.state_mutability == StateMutability.PAYABLE:
            return

        visited: Set[FunctionDefinition] = {func}
        queue: Deque[FunctionDefinition] = deque([func])

        while len(queue) > 0:
            func = queue.pop()
            for f in chain(func.base_functions, func.child_functions):
                if isinstance(f, FunctionDefinition) and f not in visited:
                    visited.add(f)
                    queue.append(f)

            for ref in func.references:
                if isinstance(ref, IdentifierPathPart):
                    func_call = ref.underlying_node
                elif isinstance(ref, ExternalReference):
                    func_call = ref.inline_assembly
                else:
                    func_call = ref
                while func_call is not None:
                    if (
                        isinstance(func_call, FunctionCall)
                        and func_call.function_called == func
                    ):
                        break
                    func_call = func_call.parent

                if func_call is None:
                    continue

                other_func = func_call
                while other_func is not None:
                    if isinstance(other_func, FunctionDefinition):
                        break
                    other_func = other_func.parent

                if other_func is None or other_func in visited:
                    continue

                if other_func.state_mutability == StateMutability.PAYABLE:
                    return

                queue.append(other_func)

        self._detections.append(
            DetectorResult(
                node,
                "`msg.value` used in a function that is never called from a payable function",
            )
        )



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Iterator, List, Tuple, Union

if TYPE_CHECKING:
    from ..declaration.error_definition import ErrorDefinition
    from ..declaration.event_definition import EventDefinition
    from ..declaration.function_definition import FunctionDefinition
    from ..declaration.modifier_definition import ModifierDefinition
    from ..type_name.function_type_name import FunctionTypeName
    from .try_catch_clause import TryCatchClause

from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.ir.declaration.variable_declaration import VariableDeclaration
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import SolcParameterList

logger = logging.getLogger(__name__)


class ParameterList(SolidityAbc):
    """
    !!! example
        A parameter list can be used:

        - in an [ErrorDefinition][woke.ast.ir.declaration.error_definition.ErrorDefinition]:
            - `:::solidity (uint requested, uint available)` in line 2,
        - in an [EventDefinition][woke.ast.ir.declaration.event_definition.EventDefinition]:
            - `:::solidity (address indexed previousOwner, address indexed newOwner)` in line 3,
        - in a [FunctionDefinition][woke.ast.ir.declaration.function_definition.FunctionDefinition]:
            - `:::solidity (uint a, uint b)` in line 12 as function parameters,
            - `:::solidity (uint256)` in line 12 as function return parameters,
        - in a [FunctionTypeName][woke.ast.ir.type_name.function_type_name.FunctionTypeName]:
            - `:::solidity (string memory, uint)` in line 5 as function type name parameters,
            - `:::solidity (bool)` in line 5 as function type name return parameters,
        - in a [ModifierDefinition][woke.ast.ir.declaration.modifier_definition.ModifierDefinition]:
            - `:::solidity (uint x)` in line 7,
        - in a [TryCatchClause][woke.ast.ir.meta.try_catch_clause.TryCatchClause]:
            - `:::solidity (bool success)` in line 17 as try clause parameters,
            - `:::solidity (string memory reason)` in line 19 as catch clause parameters.

        ```solidity linenums="1"
        contract C {
            error InsufficientBalance(uint requested, uint available);
            event OwnershipTransferred(address indexed previousOwner, address indexed newOwner);

            function (string memory, uint) external returns (bool) externalFunction;

            modifier onlyPositive(uint x) {
                require(x > 0, "x must be positive");
                _;
            }

            function add(uint a, uint b) onlyPositive(a) public pure returns (uint256) {
                return a + b;
            }

            function callExternalFunction() public {
                try externalFunction("abc", 123) returns (bool success) {
                    // ...
                } catch Error(string memory reason) {
                    // ...
                }
            }
        }
        ```
    """

    _ast_node: SolcParameterList
    _parent: Union[
        ErrorDefinition,
        EventDefinition,
        FunctionDefinition,
        FunctionTypeName,
        ModifierDefinition,
        TryCatchClause,
    ]

    _parameters: List[VariableDeclaration]

    def __init__(
        self, init: IrInitTuple, parameter_list: SolcParameterList, parent: SolidityAbc
    ):
        super().__init__(init, parameter_list, parent)

        self._parameters = []
        for parameter in parameter_list.parameters:
            self._parameters.append(VariableDeclaration(init, parameter, self))

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        for parameter in self._parameters:
            yield from parameter

    @property
    def parent(
        self,
    ) -> Union[
        ErrorDefinition,
        EventDefinition,
        FunctionDefinition,
        FunctionTypeName,
        ModifierDefinition,
        TryCatchClause,
    ]:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def parameters(self) -> Tuple[VariableDeclaration, ...]:
        """
        Can be empty.
        Returns:
            Variable declarations of the parameter list.
        """
        return tuple(self._parameters)



from __future__ import annotations

import re
from functools import lru_cache, partial
from pathlib import Path
from typing import TYPE_CHECKING, FrozenSet, Iterator, Optional, Set, Tuple, Union

from intervaltree import IntervalTree

from woke.ast.enums import (
    InlineAssemblyEvmVersion,
    InlineAssemblyFlag,
    InlineAssemblySuffix,
    ModifiesStateFlag,
)
from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.ir.declaration.abc import DeclarationAbc
from woke.ast.ir.reference_resolver import CallbackParams, ReferenceResolver
from woke.ast.ir.statement.abc import StatementAbc
from woke.ast.ir.utils import IrInitTuple
from woke.ast.ir.yul.block import Block as YulBlock
from woke.ast.nodes import AstNodeId, ExternalReferenceModel, SolcInlineAssembly

from ..yul.identifier import Identifier

if TYPE_CHECKING:
    from .block import Block
    from .do_while_statement import DoWhileStatement
    from .for_statement import ForStatement
    from .if_statement import IfStatement
    from .unchecked_block import UncheckedBlock
    from .while_statement import WhileStatement


IDENTIFIER_RE = re.compile(r"^[a-zA-Z$_][a-zA-Z0-9$_]*".encode("utf-8"))


class ExternalReference:
    """
    Reference from an inline assembly block to a Solidity declaration.
    !!! warning
        This is not an IR node, but a helper class for [InlineAssembly][woke.ast.ir.statement.inline_assembly.InlineAssembly].
        Since this is not an IR node, there must still be a Yul IR node (Yul [Identifier][woke.ast.ir.yul.identifier.Identifier]) in the source code that represents the identifier.
    """

    _inline_assembly: InlineAssembly
    _external_reference_model: ExternalReferenceModel
    _reference_resolver: ReferenceResolver
    _cu_hash: bytes
    _file: Path
    _source: bytes

    _referenced_declaration_id: AstNodeId
    _value_size: int
    _suffix: Optional[InlineAssemblySuffix]
    _yul_identifier: Optional[Identifier]

    def __init__(
        self,
        inline_assembly: InlineAssembly,
        init: IrInitTuple,
        external_reference_model: ExternalReferenceModel,
    ):
        self._inline_assembly = inline_assembly
        self._external_reference_model = external_reference_model
        self._reference_resolver = init.reference_resolver
        self._cu_hash = init.cu.hash
        self._file = init.file
        self._source = init.source[self.byte_location[0] : self.byte_location[1]]

        self._referenced_declaration_id = external_reference_model.declaration
        assert self._referenced_declaration_id >= 0
        self._value_size = external_reference_model.value_size
        self._suffix = external_reference_model.suffix
        self._yul_identifier = None

        if external_reference_model.is_offset:
            self._suffix = InlineAssemblySuffix.OFFSET
        elif external_reference_model.is_slot:
            self._suffix = InlineAssemblySuffix.SLOT

        self._reference_resolver.register_post_process_callback(self._post_process)

    def _post_process(self, callback_params: CallbackParams):
        referenced_declaration = self.referenced_declaration
        referenced_declaration.register_reference(self)
        self._reference_resolver.register_destroy_callback(
            self._file, partial(self._destroy, referenced_declaration)
        )
        interval_tree = callback_params.interval_trees[self._file]
        start, end = self.byte_location
        nodes = interval_tree[start:end]
        node = next(node for node in nodes if node.begin == start and node.end == end)
        assert isinstance(
            node.data, Identifier
        ), f"Expected Identifier, got {type(node.data)}"
        self._yul_identifier = node.data
        self._yul_identifier._external_reference = self

    def _destroy(self, referenced_declaration: DeclarationAbc) -> None:
        referenced_declaration.unregister_reference(self)

    @property
    def file(self) -> Path:
        """
        Returns:
            Absolute path to the file containing the inline assembly block.
        """
        return self._file

    @property
    def byte_location(self) -> Tuple[int, int]:
        """
        Returns:
            Byte offsets (start and end) of the external reference in the source file.
        """
        return (
            self._external_reference_model.src.byte_offset,
            self._external_reference_model.src.byte_offset
            + self._external_reference_model.src.byte_length,
        )

    @property
    @lru_cache(maxsize=2048)
    def identifier_byte_location(self) -> Tuple[int, int]:
        """
        !!! example
            Returns the byte location of `stateVar` in line 6, not `stateVar.slot`:
            ```solidity linenums="1"
            contract Foo {
                uint stateVar;

                function f() public pure {
                    assembly {
                        let x := stateVar.slot
                    }
                }
            }
            ```
        Returns:
            Byte offsets (start and end) of the identifier representing the external reference in the source file.
        """
        match = IDENTIFIER_RE.match(self._source)
        assert match
        start = self.byte_location[0] + match.start()
        end = self.byte_location[0] + match.end()
        return start, end

    @property
    def referenced_declaration(self) -> DeclarationAbc:
        """
        Returns:
            Solidity declaration referenced by this external reference.
        """
        node = self._reference_resolver.resolve_node(
            self._referenced_declaration_id, self._cu_hash
        )
        assert isinstance(node, DeclarationAbc)
        return node

    @property
    def inline_assembly(self) -> InlineAssembly:
        """
        Returns:
            Inline assembly block this external references belongs to.
        """
        return self._inline_assembly

    @property
    def yul_identifier(self) -> Identifier:
        """
        Returns:
            Yul Identifier node representing this external reference.
        """
        assert isinstance(self._yul_identifier, Identifier)
        return self._yul_identifier

    @property
    def value_size(self) -> int:
        # TODO document this?
        return self._value_size

    @property
    def suffix(self) -> Optional[InlineAssemblySuffix]:
        """
        Returns:
            Suffix of the external reference, if any.
        """
        return self._suffix


class InlineAssembly(StatementAbc):
    """
    Inline assembly block in Solidity.
    !!! example
        ```solidity
        function f() public pure {
            assembly {
                let x := 1
                let y := 2
                let z := add(x, y)
            }
        }
        ```
    """

    _ast_node: SolcInlineAssembly
    _parent: Union[
        Block,
        DoWhileStatement,
        ForStatement,
        IfStatement,
        UncheckedBlock,
        WhileStatement,
    ]

    _yul_block: YulBlock
    _evm_version: InlineAssemblyEvmVersion
    _external_references: IntervalTree
    _flags: Set[InlineAssemblyFlag]

    def __init__(
        self,
        init: IrInitTuple,
        inline_assembly: SolcInlineAssembly,
        parent: SolidityAbc,
    ):
        super().__init__(init, inline_assembly, parent)
        self._yul_block = YulBlock(init, inline_assembly.ast, self)
        self._evm_version = inline_assembly.evm_version
        self._external_references = IntervalTree()
        self._flags = set()
        if inline_assembly.flags is not None:
            for flag in inline_assembly.flags:
                self._flags.add(InlineAssemblyFlag(flag))
        for external_reference in inline_assembly.external_references:
            start = external_reference.src.byte_offset
            end = start + external_reference.src.byte_length
            self._external_references[start:end] = ExternalReference(
                self, init, external_reference
            )

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        yield from self._yul_block

    @property
    def parent(
        self,
    ) -> Union[
        Block,
        DoWhileStatement,
        ForStatement,
        IfStatement,
        UncheckedBlock,
        WhileStatement,
    ]:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def yul_block(self) -> YulBlock:
        """
        Returns:
            Yul block containing Yul IR nodes ([YulAbc][woke.ast.ir.yul.abc.YulAbc]).
        """
        return self._yul_block

    @property
    def evm_version(self) -> InlineAssemblyEvmVersion:
        """
        Depends on the version of the `solc` compiler used to compile the contract.
        Returns:
            EVM version used for the inline assembly block.
        """
        return self._evm_version

    @property
    def flags(self) -> FrozenSet[InlineAssemblyFlag]:
        """
        !!! example
            ```solidity
            function f() public pure {
                assembly ("memory-safe") {
                    let x := 1
                    let y := 2
                    let z := add(x, y)
                }
            }
            ```
        Returns:
            Flags decorating the inline assembly block.
        """
        return frozenset(self._flags)

    @property
    def external_references(self) -> Tuple[ExternalReference, ...]:
        """
        Returns:
            External references in the inline assembly block.
        """
        return tuple(interval.data for interval in self._external_references)

    def external_reference_at(self, byte_offset: int) -> Optional[ExternalReference]:
        """
        Args:
            byte_offset: Byte offset in the source file.
        Returns:
            External reference at the given byte offset, if any.
        """
        intervals = self._external_references.at(byte_offset)
        assert len(intervals) <= 1
        if len(intervals) == 0:
            return None
        return intervals.pop().data

    @property
    @lru_cache(maxsize=2048)
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        return self.yul_block.modifies_state



import copy
import logging
from collections import deque
from typing import Deque, Dict, List, Optional, Set, Tuple, Union

import networkx as nx

import woke.ast.types as types
from woke.analysis.cfg import CfgBlock
from woke.ast.enums import FunctionCallKind, GlobalSymbolsEnum
from woke.ast.ir.declaration.contract_definition import ContractDefinition
from woke.ast.ir.declaration.function_definition import FunctionDefinition
from woke.ast.ir.declaration.variable_declaration import VariableDeclaration
from woke.ast.ir.expression.abc import ExpressionAbc
from woke.ast.ir.expression.assignment import Assignment
from woke.ast.ir.expression.elementary_type_name_expression import (
    ElementaryTypeNameExpression,
)
from woke.ast.ir.expression.function_call import FunctionCall
from woke.ast.ir.expression.function_call_options import FunctionCallOptions
from woke.ast.ir.expression.identifier import Identifier
from woke.ast.ir.expression.member_access import MemberAccess
from woke.ast.ir.statement.expression_statement import ExpressionStatement
from woke.ast.ir.statement.inline_assembly import InlineAssembly
from woke.ast.ir.statement.return_statement import Return
from woke.ast.ir.yul.abc import YulAbc
from woke.ast.ir.yul.assignment import Assignment as YulAssignment
from woke.ast.ir.yul.identifier import Identifier as YulIdentifier

logger = logging.getLogger(__name__)


def pair_function_call_arguments(
    definition: FunctionDefinition, call: FunctionCall
) -> Tuple[Tuple[VariableDeclaration, ExpressionAbc], ...]:
    if len(definition.parameters.parameters) == len(call.arguments):
        return tuple(zip(definition.parameters.parameters, call.arguments))
    elif len(definition.parameters.parameters) == len(call.arguments) + 1:
        # using for
        node = call.expression
        if isinstance(node, FunctionCallOptions):
            node = node.expression
        if isinstance(node, MemberAccess):
            node = node.expression
        return ((definition.parameters.parameters[0], node),) + tuple(
            zip(definition.parameters.parameters[1:], call.arguments)
        )
    else:
        raise ValueError(
            f"{definition.name} has {len(definition.parameters.parameters)} parameters but called with {len(call.arguments)} arguments"
        )


def get_all_base_and_child_functions(
    func: Union[FunctionDefinition, VariableDeclaration]
) -> Set[Union[FunctionDefinition, VariableDeclaration]]:
    ret = {func}
    queue: Deque[Union[FunctionDefinition, VariableDeclaration]] = deque([func])

    while len(queue) > 0:
        func = queue.popleft()

        if isinstance(func, VariableDeclaration):
            for base in func.base_functions:
                if base not in ret:
                    ret.add(base)
                    queue.append(base)
        elif isinstance(func, FunctionDefinition):
            for base in func.base_functions:
                if base not in ret:
                    ret.add(base)
                    queue.append(base)
            for child in func.child_functions:
                if child not in ret:
                    ret.add(child)
                    queue.append(child)
    return ret


def expression_is_global_symbol(
    expr: ExpressionAbc, global_symbol: GlobalSymbolsEnum
) -> bool:
    if isinstance(expr, MemberAccess):
        if expr.referenced_declaration == global_symbol:
            logger.debug(f"Expression is {global_symbol}: {expr.source}")
            return True
    elif (
        isinstance(expr, FunctionCall)
        and expr.kind == FunctionCallKind.TYPE_CONVERSION
        and len(expr.arguments) == 1
    ):
        # payable(msg.sender) for example
        conversion = expr.expression
        if isinstance(conversion, ElementaryTypeNameExpression):
            t = conversion.type
            if isinstance(t, types.Type) and isinstance(t.actual_type, types.Address):
                return expression_is_global_symbol(expr.arguments[0], global_symbol)
    elif isinstance(expr, FunctionCall) and isinstance(
        expr.function_called, FunctionDefinition
    ):
        func = expr.function_called
        assert isinstance(func, FunctionDefinition)
        if len(func.return_parameters.parameters) == 1:
            returns = []
            if func.body is None:
                return False

            for statement in func.body.statements_iter():
                if isinstance(statement, Return):
                    returns.append(statement)
            if all(
                expression_is_global_symbol(return_.expression, global_symbol)
                for return_ in returns
                if return_.expression is not None
            ):
                logger.debug(f"Expression is {global_symbol}: {expr.source}")
                return True

    logger.debug(f"Expression is NOT {global_symbol}: {expr.source}")
    return False


def get_variable_declarations_from_expression(
    expr: ExpressionAbc,
) -> Tuple[Optional[VariableDeclaration], ...]:
    if isinstance(expr, Identifier):
        ref = expr.referenced_declaration
        if isinstance(ref, VariableDeclaration):
            return (ref,)
    elif isinstance(expr, MemberAccess):
        ref = expr.referenced_declaration
        if isinstance(ref, VariableDeclaration):
            return get_variable_declarations_from_expression(expr.expression) + (ref,)
    elif isinstance(expr, FunctionCall) and isinstance(
        expr.function_called, (FunctionDefinition, VariableDeclaration)
    ):
        func = expr.function_called

        func_expr = expr.expression
        if isinstance(func_expr, Identifier):
            ret: Tuple[Optional[VariableDeclaration], ...] = tuple()
        elif isinstance(func_expr, MemberAccess):
            ret = get_variable_declarations_from_expression(func_expr.expression)
        else:
            return (None,)

        if isinstance(func, VariableDeclaration):
            return ret + (func,)
        elif isinstance(func, FunctionDefinition):
            if func.body is None:
                implementations = get_function_implementations(func)
                if len(implementations) == 1:
                    if isinstance(implementations[0], VariableDeclaration):
                        return ret + (implementations[0],)
                    elif isinstance(implementations[0], FunctionDefinition):
                        func = implementations[0]
                        assert func.body is not None
                    else:
                        return (None,)
                else:
                    return (None,)

            returns = []
            for statement in func.body.statements_iter():
                if isinstance(statement, Return):
                    returns.append(statement)
            if len(returns) == 1 and returns[0].expression is not None:
                return ret + get_variable_declarations_from_expression(
                    returns[0].expression
                )

    return (None,)


def get_function_implementations(
    function: FunctionDefinition,
) -> Tuple[Union[FunctionDefinition, VariableDeclaration], ...]:
    ret = set()
    visited: Set[Union[FunctionDefinition, VariableDeclaration]] = {function}

    queue = deque([function])
    while len(queue) > 0:
        func = queue.popleft()
        if func.implemented:
            ret.add(func)

        for f in func.child_functions:
            if f not in visited:
                visited.add(f)
                if isinstance(f, VariableDeclaration):
                    ret.add(f)
                elif isinstance(f, FunctionDefinition):
                    queue.append(f)
    return tuple(ret)


def get_function_definition_from_expression(
    expr: ExpressionAbc,
) -> Optional[FunctionDefinition]:
    fn_dec = expr
    while (
        fn_dec.parent
        and not isinstance(fn_dec, FunctionDefinition)
        and not isinstance(fn_dec, ContractDefinition)
    ):
        fn_dec = fn_dec.parent
    if (
        fn_dec is not None
        and isinstance(fn_dec, FunctionDefinition)
        and fn_dec.body is not None
    ):
        return fn_dec
    return None


def check_all_return_params_set(
    params: Tuple[VariableDeclaration],
    graph: nx.DiGraph,
    start_block: CfgBlock,
    end_block: CfgBlock,
) -> Tuple[
    bool, List[Dict[Tuple[Optional[VariableDeclaration]], Union[ExpressionAbc, YulAbc]]]
]:
    """
    Checks if all return parameters are set
    """
    out = []
    visited = set()
    all_set = _check_all_return_params_set(
        params, {}, set(), graph, start_block, end_block, out, visited
    )
    return all_set, out


def _check_all_return_params_set(
    params: Tuple[VariableDeclaration],
    solved_params: Dict[
        Tuple[Optional[VariableDeclaration]], Union[ExpressionAbc, YulAbc]
    ],
    solved_declarations: Set[VariableDeclaration],
    graph: nx.DiGraph,
    start_block: CfgBlock,
    end_block: CfgBlock,
    out: List[Dict[Tuple[Optional[VariableDeclaration]], Union[ExpressionAbc, YulAbc]]],
    visited: Set[CfgBlock],
) -> bool:
    if start_block in visited:
        return False
    visited.add(start_block)
    if len(solved_declarations) == len(params):
        out.append(solved_params)
        return True
    if start_block == end_block:
        out.append(solved_params)
        return False

    for stmt in reversed(start_block.statements):
        if isinstance(stmt, ExpressionStatement) and isinstance(
            stmt.expression, Assignment
        ):
            assigned_vars = []
            assigned_params = set()
            for vars in stmt.expression.assigned_variables:
                if vars is None:
                    continue
                for decls in vars:
                    for vd in decls:
                        if (
                            isinstance(vd, VariableDeclaration)
                            and vd in params
                            and vd not in solved_declarations
                        ):
                            assigned_vars.append(vd)
                            assigned_params.add(vd)
                        else:
                            assigned_vars.append(None)
            if len(assigned_params) > 0:
                solved_declarations.update(assigned_params)
                solved_params[tuple(assigned_vars)] = stmt.expression.right_expression
        elif isinstance(stmt, InlineAssembly) and stmt.yul_block is not None:
            for yul_stmt in stmt.yul_block.statements:
                if isinstance(yul_stmt, YulAssignment):
                    assigned_vars = []
                    assigned_params = set()
                    for var in yul_stmt.variable_names:
                        if (
                            isinstance(var, YulIdentifier)
                            and var.external_reference is not None
                            and isinstance(
                                var.external_reference.referenced_declaration,
                                VariableDeclaration,
                            )
                            and var.external_reference.referenced_declaration in params
                        ):
                            assigned_vars.append(
                                var.external_reference.referenced_declaration
                            )
                            assigned_params.add(
                                var.external_reference.referenced_declaration
                            )
                        else:
                            assigned_vars.append(None)
                    if len(assigned_params) > 0:
                        solved_declarations.update(assigned_params)
                        solved_params[tuple(assigned_vars)] = yul_stmt.value

    all_set = True
    for next_start in graph.predecessors(start_block):
        if next_start in visited:
            continue
        if not _check_all_return_params_set(
            params,
            copy.copy(solved_params),
            copy.copy(solved_declarations),
            graph,
            next_start,
            end_block,
            out,
            visited,
        ):
            all_set = False
    return all_set



import logging
import re
from functools import lru_cache, partial
from typing import Iterator, Optional, Set, Tuple, Union

from woke.ast.enums import GlobalSymbolsEnum, ModifiesStateFlag
from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.ir.declaration.abc import DeclarationAbc
from woke.ast.ir.declaration.enum_definition import EnumDefinition
from woke.ast.ir.declaration.variable_declaration import VariableDeclaration
from woke.ast.ir.expression.abc import ExpressionAbc
from woke.ast.ir.expression.identifier import Identifier
from woke.ast.ir.meta.import_directive import ImportDirective
from woke.ast.ir.meta.source_unit import SourceUnit
from woke.ast.ir.reference_resolver import CallbackParams
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import AstNodeId, SolcMemberAccess
from woke.ast.types import (
    Address,
    Array,
    Bytes,
    FixedBytes,
    Function,
    Magic,
    MagicTypeKind,
    String,
    Type,
    UserDefinedValueType,
)

MEMBER_RE = re.compile(r"\s*.\s*(?P<member>.+)".encode("utf-8"))


logger = logging.getLogger(__name__)


class MemberAccess(ExpressionAbc):
    """
    TBD
    """

    _ast_node: SolcMemberAccess
    _parent: SolidityAbc  # TODO: make this more specific

    _expression: ExpressionAbc
    _member_name: str
    _referenced_declaration_id: Optional[AstNodeId]

    def __init__(
        self, init: IrInitTuple, member_access: SolcMemberAccess, parent: SolidityAbc
    ):
        super().__init__(init, member_access, parent)
        self._expression = ExpressionAbc.from_ast(init, member_access.expression, self)
        assert self._expression.byte_location[0] == self.byte_location[0]
        assert self._expression.byte_location[1] < self.byte_location[1]

        self._member_name = member_access.member_name
        self._referenced_declaration_id = member_access.referenced_declaration

        self._reference_resolver.register_post_process_callback(
            self._post_process, priority=-1
        )

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        yield from self._expression

    def _post_process(self, callback_params: CallbackParams):
        # workaround for enum value bug in Solidity versions prior to 0.8.2
        if self._referenced_declaration_id is None:
            if isinstance(self._expression, Identifier) or (
                isinstance(self._expression, MemberAccess)
                and self._expression._referenced_declaration_id is not None
            ):
                referenced_declaration = self._expression.referenced_declaration
                if isinstance(referenced_declaration, EnumDefinition):
                    for enum_value in referenced_declaration.values:
                        if enum_value.name == self._member_name:
                            node_path_order = (
                                self._reference_resolver.get_node_path_order(
                                    AstNodeId(enum_value.ast_node_id),
                                    enum_value.cu_hash,
                                )
                            )
                            this_cu_id = self._reference_resolver.get_ast_id_from_cu_node_path_order(
                                node_path_order, self.cu_hash
                            )
                            self._referenced_declaration_id = this_cu_id
                            break

        if self._referenced_declaration_id is None:
            expr_type = self.expression.type

            if isinstance(expr_type, Address):
                if self.member_name == "balance":
                    self._referenced_declaration_id = AstNodeId(
                        GlobalSymbolsEnum.ADDRESS_BALANCE
                    )
                elif self.member_name == "code":
                    self._referenced_declaration_id = AstNodeId(
                        GlobalSymbolsEnum.ADDRESS_CODE
                    )
                elif self.member_name == "codehash":
                    self._referenced_declaration_id = AstNodeId(
                        GlobalSymbolsEnum.ADDRESS_CODEHASH
                    )
                elif self.member_name == "transfer":
                    self._referenced_declaration_id = AstNodeId(
                        GlobalSymbolsEnum.ADDRESS_TRANSFER
                    )
                elif self.member_name == "send":
                    self._referenced_declaration_id = AstNodeId(
                        GlobalSymbolsEnum.ADDRESS_SEND
                    )
                elif self.member_name == "call":
                    self._referenced_declaration_id = AstNodeId(
                        GlobalSymbolsEnum.ADDRESS_CALL
                    )
                elif self.member_name == "delegatecall":
                    self._referenced_declaration_id = AstNodeId(
                        GlobalSymbolsEnum.ADDRESS_DELEGATECALL
                    )
                elif self.member_name == "staticcall":
                    self._referenced_declaration_id = AstNodeId(
                        GlobalSymbolsEnum.ADDRESS_STATICCALL
                    )
                else:
                    assert False, f"Unknown address member: {self.member_name}"
            elif isinstance(expr_type, Array):
                if self.member_name == "length":
                    self._referenced_declaration_id = AstNodeId(
                        GlobalSymbolsEnum.ARRAY_LENGTH
                    )
                elif self.member_name == "push":
                    self._referenced_declaration_id = AstNodeId(
                        GlobalSymbolsEnum.ARRAY_PUSH
                    )
                elif self.member_name == "pop":
                    self._referenced_declaration_id = AstNodeId(
                        GlobalSymbolsEnum.ARRAY_POP
                    )
                else:
                    assert False, f"Unknown array member: {self.member_name}"
            elif isinstance(expr_type, (Bytes, FixedBytes)):
                if self.member_name == "length":
                    self._referenced_declaration_id = AstNodeId(
                        GlobalSymbolsEnum.BYTES_LENGTH
                    )
                elif self.member_name == "push":
                    self._referenced_declaration_id = AstNodeId(
                        GlobalSymbolsEnum.BYTES_PUSH
                    )
                else:
                    assert False, f"Unknown bytes member: {self.member_name}"
            elif isinstance(expr_type, Function):
                if self.member_name == "selector":
                    self._referenced_declaration_id = AstNodeId(
                        GlobalSymbolsEnum.FUNCTION_SELECTOR
                    )
                elif self.member_name == "value":
                    self._referenced_declaration_id = AstNodeId(
                        GlobalSymbolsEnum.FUNCTION_VALUE
                    )
                elif self.member_name == "gas":
                    self._referenced_declaration_id = AstNodeId(
                        GlobalSymbolsEnum.FUNCTION_GAS
                    )
                elif self.member_name == "address":
                    self._referenced_declaration_id = AstNodeId(
                        GlobalSymbolsEnum.FUNCTION_ADDRESS
                    )
                else:
                    assert False, f"Unknown function member: {self.member_name}"
            elif isinstance(expr_type, Magic):
                if expr_type.kind == MagicTypeKind.BLOCK:
                    if self.member_name == "basefee":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.BLOCK_BASEFEE
                        )
                    elif self.member_name == "chainid":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.BLOCK_CHAINID
                        )
                    elif self.member_name == "coinbase":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.BLOCK_COINBASE
                        )
                    elif self.member_name == "difficulty":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.BLOCK_DIFFICULTY
                        )
                    elif self.member_name == "gaslimit":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.BLOCK_GASLIMIT
                        )
                    elif self.member_name == "number":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.BLOCK_NUMBER
                        )
                    elif self.member_name == "timestamp":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.BLOCK_TIMESTAMP
                        )
                    elif self.member_name == "prevrandao":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.BLOCK_PREVRANDAO
                        )
                    else:
                        assert False, f"Unknown block member {self.member_name}"
                elif expr_type.kind == MagicTypeKind.MESSAGE:
                    if self.member_name == "data":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.MSG_DATA
                        )
                    elif self.member_name == "sender":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.MSG_SENDER
                        )
                    elif self.member_name == "sig":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.MSG_SIG
                        )
                    elif self.member_name == "value":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.MSG_VALUE
                        )
                    else:
                        assert False, f"Unknown msg member {self.member_name}"
                elif expr_type.kind == MagicTypeKind.TRANSACTION:
                    if self.member_name == "gasprice":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.TX_GASPRICE
                        )
                    elif self.member_name == "origin":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.TX_ORIGIN
                        )
                    else:
                        assert False, f"Unknown tx member {self.member_name}"
                elif expr_type.kind == MagicTypeKind.ABI:
                    if self.member_name == "decode":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.ABI_DECODE
                        )
                    elif self.member_name == "encode":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.ABI_ENCODE
                        )
                    elif self.member_name == "encodePacked":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.ABI_ENCODE_PACKED
                        )
                    elif self.member_name == "encodeWithSelector":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.ABI_ENCODE_WITH_SELECTOR
                        )
                    elif self.member_name == "encodeWithSignature":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.ABI_ENCODE_WITH_SIGNATURE
                        )
                    elif self.member_name == "encodeCall":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.ABI_ENCODE_CALL
                        )
                    else:
                        assert False, f"Unknown abi member {self.member_name}"
                elif expr_type.kind == MagicTypeKind.META_TYPE:
                    if self.member_name == "name":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.TYPE_NAME
                        )
                    elif self.member_name == "creationCode":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.TYPE_CREATION_CODE
                        )
                    elif self.member_name == "runtimeCode":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.TYPE_RUNTIME_CODE
                        )
                    elif self.member_name == "interfaceId":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.TYPE_INTERFACE_ID
                        )
                    elif self.member_name == "min":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.TYPE_MIN
                        )
                    elif self.member_name == "max":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.TYPE_MAX
                        )
                    else:
                        assert False, f"Unknown type member {self.member_name}"
            elif isinstance(expr_type, Type):
                if isinstance(expr_type.actual_type, Bytes):
                    if self.member_name == "concat":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.BYTES_CONCAT
                        )
                    else:
                        assert False, f"Unknown bytes member {self.member_name}"
                elif isinstance(expr_type.actual_type, String):
                    if self.member_name == "concat":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.STRING_CONCAT
                        )
                    else:
                        assert False, f"Unknown string member {self.member_name}"
                elif isinstance(expr_type.actual_type, UserDefinedValueType):
                    if self.member_name == "wrap":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.USER_DEFINED_VALUE_TYPE_WRAP
                        )
                    elif self.member_name == "unwrap":
                        self._referenced_declaration_id = AstNodeId(
                            GlobalSymbolsEnum.USER_DEFINED_VALUE_TYPE_UNWRAP
                        )
                    else:
                        assert (
                            False
                        ), f"Unknown user defined value type member {self.member_name}"
                else:
                    assert (
                        False
                    ), f"Unknown type member {self.member_name} {expr_type.actual_type}"

        assert (
            self._referenced_declaration_id is not None
        ), f"Unknown member {self.member_name}"

        if self._referenced_declaration_id < 0:
            global_symbol = GlobalSymbolsEnum(self._referenced_declaration_id)
            self._reference_resolver.register_global_symbol_reference(
                global_symbol, self
            )
            self._reference_resolver.register_destroy_callback(
                self.file, partial(self._destroy, global_symbol)
            )
        else:
            node = self._reference_resolver.resolve_node(
                self._referenced_declaration_id, self._cu_hash
            )

            if isinstance(node, DeclarationAbc):
                node.register_reference(self)
                self._reference_resolver.register_destroy_callback(
                    self.file, partial(self._destroy, node)
                )
            elif isinstance(node, ImportDirective):
                # make this node to reference the source unit directly
                assert node.unit_alias is not None
                source_unit = callback_params.source_units[node.imported_file]
                node_path_order = self._reference_resolver.get_node_path_order(
                    AstNodeId(source_unit.ast_node_id),
                    source_unit.cu_hash,
                )
                self._referenced_declaration_id = (
                    self._reference_resolver.get_ast_id_from_cu_node_path_order(
                        node_path_order, self.cu_hash
                    )
                )
            else:
                raise TypeError(f"Unexpected type: {type(node)}")

    def _destroy(
        self, referenced_declaration: Union[GlobalSymbolsEnum, DeclarationAbc]
    ) -> None:
        if isinstance(referenced_declaration, GlobalSymbolsEnum):
            self._reference_resolver.unregister_global_symbol_reference(
                referenced_declaration, self
            )
        elif isinstance(referenced_declaration, DeclarationAbc):
            referenced_declaration.unregister_reference(self)
        else:
            raise TypeError(f"Unexpected type: {type(referenced_declaration)}")

    @property
    def parent(self) -> SolidityAbc:
        return self._parent

    @property
    def expression(self) -> ExpressionAbc:
        return self._expression

    @property
    def member_name(self) -> str:
        return self._member_name

    @property
    @lru_cache(maxsize=2048)
    def member_byte_location(self) -> Tuple[int, int]:
        relative_expression_end = (
            self._expression.byte_location[1] - self.byte_location[0]
        )
        match = MEMBER_RE.match(self._source[relative_expression_end:])
        assert match
        return self._expression.byte_location[1] + match.start(
            "member"
        ), self._expression.byte_location[1] + match.end("member")

    @property
    def referenced_declaration(
        self,
    ) -> Union[DeclarationAbc, GlobalSymbolsEnum, SourceUnit]:
        assert self._referenced_declaration_id is not None
        if self._referenced_declaration_id < 0:
            return GlobalSymbolsEnum(self._referenced_declaration_id)

        node = self._reference_resolver.resolve_node(
            self._referenced_declaration_id, self._cu_hash
        )
        assert isinstance(node, (DeclarationAbc, SourceUnit))
        return node

    @property
    @lru_cache(maxsize=2048)
    def is_ref_to_state_variable(self) -> bool:
        referenced_declaration = self.referenced_declaration
        return (
            isinstance(referenced_declaration, VariableDeclaration)
            and referenced_declaration.is_state_variable
            or self.expression.is_ref_to_state_variable
        )

    @property
    @lru_cache(maxsize=2048)
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        return self.expression.modifies_state



import logging
from typing import List, Union

from woke.ast.ir.abc import IrAbc
from woke.ast.ir.declaration.function_definition import FunctionDefinition
from woke.ast.ir.declaration.modifier_definition import ModifierDefinition
from woke.ast.ir.declaration.variable_declaration import VariableDeclaration
from woke.lsp.common_structures import (
    DocumentUri,
    Location,
    LocationLink,
    PartialResultParams,
    StaticRegistrationOptions,
    TextDocumentPositionParams,
    TextDocumentRegistrationOptions,
    WorkDoneProgressOptions,
    WorkDoneProgressParams,
)
from woke.lsp.context import LspContext
from woke.lsp.utils import path_to_uri, uri_to_path

logger = logging.getLogger(__name__)


class ImplementationOptions(WorkDoneProgressOptions):
    pass


class ImplementationRegistrationOptions(
    TextDocumentRegistrationOptions, ImplementationOptions, StaticRegistrationOptions
):
    pass


class ImplementationParams(
    TextDocumentPositionParams, WorkDoneProgressParams, PartialResultParams
):
    pass


async def implementation(
    context: LspContext, params: ImplementationParams
) -> Union[Location, List[Location], List[LocationLink], None]:
    logger.debug(
        f"Go to implementation for file {params.text_document.uri} at position {params.position} requested"
    )
    await context.compiler.output_ready.wait()

    path = uri_to_path(params.text_document.uri).resolve()
    if path not in context.compiler.interval_trees:
        return None

    tree = context.compiler.interval_trees[path]

    byte_offset = context.compiler.get_byte_offset_from_line_pos(
        path, params.position.line, params.position.character
    )
    intervals = tree.at(byte_offset)
    nodes: List[IrAbc] = [interval.data for interval in intervals]
    logger.debug(
        f"Found {len(nodes)} nodes at byte offset {byte_offset}:\n{sorted(nodes, key=lambda x: x.ast_tree_depth)}"
    )
    if len(nodes) == 0:
        return None

    node = max(nodes, key=lambda n: n.ast_tree_depth)
    logger.debug(f"Found node {node}")

    if not isinstance(node, (FunctionDefinition, ModifierDefinition)):
        return None

    if node.implemented:
        return None

    implementations = []
    if isinstance(node, FunctionDefinition):
        for child_function in node.child_functions:
            if isinstance(child_function, FunctionDefinition):
                if child_function.implemented:
                    implementations.append(
                        Location(
                            uri=DocumentUri(path_to_uri(child_function.file)),
                            range=context.compiler.get_range_from_byte_offsets(
                                child_function.file, child_function.name_location
                            ),
                        )
                    )
            elif isinstance(child_function, VariableDeclaration):
                implementations.append(
                    Location(
                        uri=DocumentUri(path_to_uri(child_function.file)),
                        range=context.compiler.get_range_from_byte_offsets(
                            child_function.file, child_function.name_location
                        ),
                    )
                )
    elif isinstance(node, ModifierDefinition):
        for child_modifier in node.child_modifiers:
            if child_modifier.implemented:
                implementations.append(
                    Location(
                        uri=DocumentUri(path_to_uri(child_modifier.file)),
                        range=context.compiler.get_range_from_byte_offsets(
                            child_modifier.file, child_modifier.name_location
                        ),
                    )
                )

    if len(implementations) == 0:
        return None
    return implementations



from __future__ import annotations

from collections import Counter, defaultdict
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    DefaultDict,
    Dict,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Union,
)

from tree_sitter import Parser
from tree_sitter_solidity import get_parser

from .common_structures import (
    CreateFilesParams,
    DeleteFilesParams,
    MessageType,
    RenameFilesParams,
)
from .document_sync import (
    DidChangeTextDocumentParams,
    DidCloseTextDocumentParams,
    DidOpenTextDocumentParams,
)
from .utils import uri_to_path

if TYPE_CHECKING:
    from .server import LspServer


ENCODING = "utf-16-le"


def _binary_search(lines: List[Tuple[bytes, int]], x: int) -> int:
    l = 0
    r = len(lines)

    while l < r:
        mid = l + (r - l) // 2
        if lines[mid][1] < x + 1:
            l = mid + 1
        else:
            r = mid

    return l - 1


class VersionedFile(NamedTuple):
    text: bytearray
    version: int


class LspParser:
    _parser: Parser
    _files: Dict[Path, VersionedFile]
    _trees: Dict[Path, Any]
    _tree_changed: DefaultDict[Path, bool]
    _line_indexes: Dict[Path, List[Tuple[bytes, int]]]
    _line_endings: Dict[Path, Optional[str]]
    _server: LspServer

    def __init__(self, server: LspServer):
        self._server = server
        self._files = {}
        self._trees = {}
        self._tree_changed = defaultdict(bool)
        self._line_indexes = {}
        self._line_endings = {}
        self._parser = get_parser()

    def __getitem__(self, item: Path) -> Any:
        if item not in self._trees:
            self._trees[item] = self._parser.parse(
                bytes(self._files[item].text),
                encoding="utf16",  # pyright: ignore reportGeneralTypeIssues
            )

        if self._tree_changed[item]:
            self._trees[item] = self._parser.parse(
                self._files[item].text,
                self._trees[item],
                encoding="utf16",  # pyright: ignore reportGeneralTypeIssues
            )
            self._tree_changed[item] = False

        return self._trees[item]

    @property
    def files(self) -> Set[Path]:
        return set(self._files.keys())

    def _setup_line_index(self, file: Path, content: str):
        # UTF-16 encoded lines with prefix length
        encoded_lines: List[Tuple[bytes, int]] = []
        prefix_sum = 0
        line_endings = Counter()
        for line in content.splitlines(keepends=True):
            encoded_line = line.encode(ENCODING)
            encoded_lines.append((encoded_line, prefix_sum))
            prefix_sum += len(encoded_line)

            if line.endswith("\r\n"):
                line_endings["\r\n"] += 1
            elif line.endswith("\n"):
                line_endings["\n"] += 1
            elif line.endswith("\r"):
                line_endings["\r"] += 1

        self._line_indexes[file] = encoded_lines
        try:
            self._line_endings[file] = line_endings.most_common(1)[0][0]
        except IndexError:
            self._line_endings[file] = None

    def _get_byte_offset_from_line_pos(self, file: Path, line: int, col: int) -> int:
        lines = self._line_indexes[file]
        if len(lines) == line:
            if line == 0:
                return col * 2
            return lines[line - 1][1] + len(lines[line - 1][0]) + col * 2

        line_bytes, prefix = lines[line]
        return prefix + col * 2

    def _get_line_pos_from_byte_offset(
        self, file: Path, offset: int
    ) -> Tuple[int, int]:
        encoded_lines = self._line_indexes[file]
        if len(encoded_lines) == 0:
            return 0, offset // 2

        line_num = _binary_search(self._line_indexes[file], offset)
        line_data, prefix_sum = encoded_lines[line_num]
        col_num = (offset - prefix_sum) // 2
        return line_num, col_num

    async def add_change(
        self,
        change: Union[
            DidOpenTextDocumentParams,
            DidChangeTextDocumentParams,
            DidCloseTextDocumentParams,
            CreateFilesParams,
            RenameFilesParams,
            DeleteFilesParams,
        ],
    ) -> None:
        if isinstance(change, DidOpenTextDocumentParams):
            path = uri_to_path(change.text_document.uri).resolve()
            self._files[path] = VersionedFile(
                bytearray(change.text_document.text.encode(ENCODING)),
                change.text_document.version,
            )
            self._setup_line_index(path, change.text_document.text)
        elif isinstance(change, DidChangeTextDocumentParams):
            path = uri_to_path(change.text_document.uri).resolve()
            for content_change in change.content_changes:
                encoded_change = content_change.text.encode(ENCODING)
                start_offset = self._get_byte_offset_from_line_pos(
                    path,
                    content_change.range.start.line,
                    content_change.range.start.character,
                )
                end_offset = self._get_byte_offset_from_line_pos(
                    path,
                    content_change.range.end.line,
                    content_change.range.end.character,
                )

                self._files[path].text[start_offset:end_offset] = bytearray(
                    encoded_change
                )
                self._setup_line_index(
                    path, self._files[path].text.decode(ENCODING)
                )  # TODO: Optimize

                if path in self._trees:
                    new_end_point = self._get_line_pos_from_byte_offset(
                        path, start_offset + len(encoded_change)
                    )
                    self._trees[path].edit(
                        start_byte=start_offset,
                        old_end_byte=end_offset,
                        new_end_byte=start_offset + len(encoded_change),
                        start_point=(
                            content_change.range.start.line,
                            content_change.range.start.character * 2,
                        ),
                        old_end_point=(
                            content_change.range.end.line,
                            content_change.range.end.character * 2,
                        ),
                        new_end_point=(new_end_point[0], new_end_point[1] * 2),
                    )

            self._files[path] = VersionedFile(
                self._files[path].text, change.text_document.version
            )
            self._tree_changed[path] = True
        elif isinstance(change, RenameFilesParams):
            for rename in change.files:
                old_file = uri_to_path(rename.old_uri)
                new_file = uri_to_path(rename.new_uri)

                if old_file in self._files:
                    self._files[new_file] = self._files.pop(old_file)
                if old_file in self._trees:
                    self._trees[new_file] = self._trees.pop(old_file)
                if old_file in self._tree_changed:
                    self._tree_changed[new_file] = self._tree_changed.pop(old_file)
                if old_file in self._line_indexes:
                    self._line_indexes[new_file] = self._line_indexes.pop(old_file)
        elif isinstance(change, DeleteFilesParams):
            for delete in change.files:
                path = uri_to_path(delete.uri)
                if path in self._files:
                    del self._files[path]
                if path in self._trees:
                    del self._trees[path]
                if path in self._tree_changed:
                    del self._tree_changed[path]
                if path in self._line_indexes:
                    del self._line_indexes[path]

    async def get_node_at_position(self, file: Path, line: int, col: int):
        """
        Returns the most nested node at the given position.
        """
        offset = self._get_byte_offset_from_line_pos(file, line, col)
        cursor = self[file].walk()
        last_node = cursor.node

        while True:
            while cursor.goto_first_child():
                if cursor.node.start_byte <= offset < cursor.node.end_byte:
                    pass
                else:
                    while cursor.goto_next_sibling():
                        if cursor.node.start_byte <= offset < cursor.node.end_byte:
                            break

            if cursor.node.start_byte <= offset < cursor.node.end_byte:
                pass
            else:
                cursor.goto_parent()

            if cursor.node == last_node:
                break
            last_node = cursor.node

        return last_node

    def get_line_ending(self, file: Path) -> Optional[str]:
        try:
            return self._line_endings[file]
        except KeyError:
            return None



from __future__ import annotations

import copy
import json
import logging
import pathlib
import re
import time
from collections import ChainMap, defaultdict
from dataclasses import asdict, dataclass, field
from itertools import chain
from typing import Any, Callable, DefaultDict, Dict, List, Optional, Set, Tuple, Union

from intervaltree import IntervalTree

from woke.ast.ir.abc import IrAbc
from woke.ast.ir.declaration.function_definition import FunctionDefinition
from woke.ast.ir.declaration.modifier_definition import ModifierDefinition
from woke.ast.ir.meta.source_unit import SourceUnit
from woke.ast.ir.reference_resolver import ReferenceResolver
from woke.ast.ir.statement.abc import StatementAbc
from woke.ast.ir.statement.block import Block
from woke.ast.ir.statement.do_while_statement import DoWhileStatement
from woke.ast.ir.statement.for_statement import ForStatement
from woke.ast.ir.statement.if_statement import IfStatement
from woke.ast.ir.statement.try_statement import TryStatement
from woke.ast.ir.statement.unchecked_block import UncheckedBlock
from woke.ast.ir.statement.while_statement import WhileStatement
from woke.ast.ir.yul.abc import YulStatementAbc
from woke.ast.ir.yul.block import Block as YulBlock
from woke.ast.ir.yul.for_loop import ForLoop as YulForLoop
from woke.ast.ir.yul.function_definition import (
    FunctionDefinition as YulFunctionDefinition,
)
from woke.ast.ir.yul.if_statement import If as YulIf
from woke.ast.ir.yul.switch import Switch as YulSwitch
from woke.cli.console import console
from woke.compiler import SolidityCompiler
from woke.compiler.build_data_model import ProjectBuild
from woke.config import WokeConfig
from woke.development.chain_interfaces import TxParams
from woke.development.core import (
    Address,
    Chain,
    get_fqn_from_address,
    get_fqn_from_creation_code,
)
from woke.development.internal import read_from_memory

logger = logging.getLogger(__name__)
logger.setLevel(logging.ERROR)


@dataclass
class SourceMapPcRecord:
    offset: Tuple[int, int]
    source_file: Optional[pathlib.Path]
    jump_type: str
    mod_depth: Optional[int]
    op: str
    argument: Optional[int]
    size: int


@dataclass(frozen=True, eq=True)
class IdePosition:
    start_line: int
    start_column: int
    end_line: int
    end_column: int

    def __iter__(self):
        yield self.start_line
        yield self.start_column
        yield self.end_line
        yield self.end_column


def _get_line_col_from_offset(
    byte_offset: int, line_index: List[Tuple[bytes, int]]
) -> Tuple[int, int]:
    line_num = _binary_search(line_index, byte_offset)
    line_data, prefix_sum = line_index[line_num]
    line_offset = byte_offset - prefix_sum
    return line_num, line_offset


@dataclass
class IdeCoverageRecord:
    ide_pos: IdePosition
    coverage_hits: int

    def __add__(self, other):
        self.coverage_hits += other.coverage_hits
        return self

    def export(self):
        """
        Exports record in a dictionary for IDE coverage
        """
        return {
            "startLine": self.ide_pos.start_line,
            "startColumn": self.ide_pos.start_column,
            "endLine": self.ide_pos.end_line,
            "endColumn": self.ide_pos.end_column,
            "coverageHits": self.coverage_hits,
        }


@dataclass
class IdeFunctionCoverageRecord(IdeCoverageRecord):
    name: str
    mod_records: Dict[IdePosition, IdeCoverageRecord]
    branch_records: Dict[IdePosition, IdeCoverageRecord]

    def __add__(self, other):
        self.coverage_hits += other.coverage_hits
        for pos, rec in other.mod_records.items():
            if pos in self.mod_records:
                self.mod_records[pos] += rec
            else:
                self.mod_records[pos] = copy.deepcopy(rec)
        for pos, rec in other.branch_records.items():
            if pos in self.branch_records:
                self.branch_records[pos] += rec
            else:
                self.branch_records[pos] = copy.deepcopy(rec)
        return self

    def export(self):
        """
        Exports record in a dictionary for IDE coverage
        """
        return {
            "name": self.name,
            "startLine": self.ide_pos.start_line,
            "startColumn": self.ide_pos.start_column,
            "endLine": self.ide_pos.end_line,
            "endColumn": self.ide_pos.end_column,
            "coverageHits": self.coverage_hits,
            "modRecords": [v.export() for v in self.mod_records.values()],
            "branchRecords": [v.export() for v in self.branch_records.values()],
        }


def _binary_search(lines: List[Tuple[bytes, int]], x: int) -> int:
    l = 0
    r = len(lines)

    while l < r:
        mid = l + (r - l) // 2
        if lines[mid][1] < x + 1:
            l = mid + 1
        else:
            r = mid

    return l - 1


def _setup_line_indexes(content: bytes) -> List[Tuple[bytes, int]]:
    tmp_lines = re.split(b"(\r?\n)", content)
    lines: List[bytes] = []
    for line in tmp_lines:
        if line in {b"\r\n", b"\n"}:
            lines[-1] += line
        else:
            lines.append(line)

    # UTF-8 encoded lines with prefix length
    encoded_lines: List[Tuple[bytes, int]] = []
    prefix_sum = 0
    for line in lines:
        encoded_lines.append((line, prefix_sum))
        prefix_sum += len(line)
    return encoded_lines


def _parse_opcodes(opcodes: str) -> List[Tuple[int, str, int, Optional[int]]]:
    pc_op_map = []
    opcodes_spl = opcodes.split(" ")

    pc = 0
    ignore = False

    for i, opcode in enumerate(opcodes_spl):
        if ignore:
            ignore = False
            continue

        if not opcode.startswith("PUSH") or opcode == "PUSH0":
            pc_op_map.append((pc, opcode, 1, None))
            pc += 1
        else:
            size = int(opcode[4:]) + 1
            pc_op_map.append((pc, opcode, size, int(opcodes_spl[i + 1], 16)))
            pc += size
            ignore = True
    return pc_op_map


def _parse_source_map(
    cu_hash: bytes,
    reference_resolver: ReferenceResolver,
    source_map: str,
    pc_op_map: List[Tuple[int, str, int, Optional[int]]],
) -> Dict[int, SourceMapPcRecord]:
    pc_map = {}
    source_map_spl = source_map.split(";")

    last_data = [-1, -1, -1, None, None]

    for i, sm_item in enumerate(source_map_spl):
        pc, op, size, argument = pc_op_map[i]
        source_spl = sm_item.split(":")
        for x in range(len(source_spl)):
            if source_spl[x] == "":
                continue
            if x < 3:
                last_data[x] = int(source_spl[x])
            else:
                last_data[x] = source_spl[x]

        source_interval = (last_data[0], last_data[0] + last_data[1], last_data[2])

        try:
            path = reference_resolver.resolve_source_file_id(
                source_interval[2], cu_hash
            )
        except KeyError:
            path = None
        pc_map[pc] = SourceMapPcRecord(
            (source_interval[0], source_interval[1]),
            path,
            last_data[3],
            last_data[4],
            op,
            argument,
            size,
        )

    return pc_map


def _construct_coverage_data(
    build: ProjectBuild, use_deployed_bytecode: bool = True
) -> Dict[str, Dict[int, SourceMapPcRecord]]:
    pc_maps = {}

    for unit_path, source_unit in build.source_units.items():
        for contract in source_unit.contracts:
            assert contract.compilation_info is not None
            assert contract.compilation_info.evm is not None

            if use_deployed_bytecode:
                bytecode = contract.compilation_info.evm.deployed_bytecode
            else:
                bytecode = contract.compilation_info.evm.bytecode

            assert bytecode is not None
            assert bytecode.opcodes is not None
            assert bytecode.source_map is not None

            opcodes = bytecode.opcodes
            source_map = bytecode.source_map

            logger.debug(f"Processing {contract.name} from {unit_path}")
            logger.debug(f"{contract.name} Opcodes {opcodes}")
            pc_op_map = _parse_opcodes(opcodes)
            logger.debug(f"{contract.name} Pc Op Map {pc_op_map}")
            pc_map = _parse_source_map(
                contract.cu_hash,
                build.reference_resolver,
                source_map,
                pc_op_map,
            )
            logger.debug(f"{contract.name} Pc Map {pc_map}")

            contract_fqn = f"{source_unit.source_unit_name}:{contract.name}"
            pc_maps[contract_fqn] = pc_map
    return pc_maps


def _merge_ide_function_coverages(
    func_covs: List[Dict[pathlib.Path, Dict[IdePosition, IdeFunctionCoverageRecord]]]
) -> Dict[pathlib.Path, Dict[IdePosition, IdeFunctionCoverageRecord]]:
    merged = {}

    for func_cov in func_covs:
        for file_path, ide_func_cov_recs in func_cov.items():
            if file_path not in merged:
                merged[file_path] = {}
            for ide_pos, func_cov_rec in ide_func_cov_recs.items():
                if ide_pos not in merged[file_path]:
                    merged[file_path][ide_pos] = copy.deepcopy(func_cov_rec)
                else:
                    merged[file_path][ide_pos] += func_cov_rec
    return merged


def export_merged_ide_coverage(
    func_covs: List[Dict[pathlib.Path, Dict[IdePosition, IdeFunctionCoverageRecord]]]
) -> Dict[str, List[Dict[str, List]]]:
    """
    Merges and exports IDE coverage records into a dictionary of paths -> funcs and their
    coverage records
    """
    merged_records = _merge_ide_function_coverages(func_covs)

    exported_coverage = {}
    for file_path, func_cov_recs in merged_records.items():
        file_path_str = str(file_path)
        if file_path not in exported_coverage:
            exported_coverage[file_path_str] = []
        for ide_pos, func_cov_rec in func_cov_recs.items():
            exported_coverage[file_path_str].append(func_cov_rec.export())
    return exported_coverage


@dataclass
class CoverageFileData:
    version: str = "1.0"
    data: Dict[str, List[Dict[str, List]]] = field(default_factory=dict)


def write_coverage(
    coverage: Dict[str, List[Dict[str, List]]], coverage_file: pathlib.Path
):
    data = CoverageFileData(data=coverage)
    with open(coverage_file, "w") as f:
        json.dump(asdict(data), f, indent=4)


class CoverageHandler:
    _pc_maps: Dict[str, Dict[int, SourceMapPcRecord]]
    _pc_maps_undeployed: Dict[str, Dict[int, SourceMapPcRecord]]
    _source_units: Dict[pathlib.Path, SourceUnit]
    _interval_trees: Dict[pathlib.Path, IntervalTree]
    _lines_index: Dict[pathlib.Path, List[Tuple[bytes, int]]]
    _statement_coverage: DefaultDict[Union[StatementAbc, YulStatementAbc], int]
    _function_coverage: DefaultDict[FunctionDefinition, int]
    _modifier_coverage: DefaultDict[ModifierDefinition, int]
    _visited_functions: Set[FunctionDefinition]
    _visited_modifiers: Set[ModifierDefinition]
    _last_statements: DefaultDict[
        Union[FunctionDefinition, ModifierDefinition],
        Optional[Tuple[Union[StatementAbc, YulStatementAbc], int]],
    ]
    _callback: Optional[Callable]

    def __init__(self, config: WokeConfig):
        compiler = SolidityCompiler(config)
        compiler.load(console=console)

        if compiler.latest_build is None or compiler.latest_build_info is None:
            raise RuntimeError(
                "Failed to load previous build. Run `woke compile` first."
            )

        self._source_units = compiler.latest_build.source_units
        self._interval_trees = compiler.latest_build.interval_trees
        self._lines_index = {}
        self._statement_coverage = defaultdict(int)
        self._function_coverage = defaultdict(int)
        self._modifier_coverage = defaultdict(int)
        self._visited_functions = set()
        self._visited_modifiers = set()
        self._last_statements = defaultdict(lambda: None)
        self._callback = None

        errored = False
        for cu in compiler.latest_build_info.compilation_units.values():
            if any(e for e in cu.errors if e.severity == "error"):
                errored = True
                break

        if errored:
            console.print(
                "[yellow]Warning: There are errors in your contracts. Coverage may be inaccurate.[/yellow]"
            )

        start = time.perf_counter()
        with console.status("[bold green]Preparing coverage data...[/]"):
            self._pc_maps = _construct_coverage_data(
                compiler.latest_build, use_deployed_bytecode=True
            )
            self._pc_maps_undeployed = _construct_coverage_data(
                compiler.latest_build, use_deployed_bytecode=False
            )
            for source_unit in compiler.latest_build.source_units.values():
                self._lines_index[source_unit.file] = _setup_line_indexes(
                    source_unit.file_source
                )

        end = time.perf_counter()
        console.log(
            f"[green]Prepared coverage data in [bold green]{end - start:.2f} s[/bold green][/]"
        )

    def set_callback(self, callback: Callable) -> None:
        self._callback = callback

    def add_coverage(
        self, params: TxParams, chain: Chain, debug_trace: Dict[str, Any]
    ) -> None:
        fqn_overrides: ChainMap[Address, Optional[str]] = ChainMap()
        # TODO process fqn overrides for tx: process txs in the same block before the given tx
        # TODO what to do with call?

        if "to" not in params or params["to"] is None:
            assert "data" in params
            bytecode = params["data"]
            try:
                contract_fqn, _ = get_fqn_from_creation_code(bytecode)
                logger.info(f"Contract {contract_fqn} was deployed")
            except ValueError:
                logger.warning(f"Failed to get contract FQN for {bytecode}")
                contract_fqn = None

            self.process_trace(
                contract_fqn, debug_trace, chain, fqn_overrides, is_from_deployment=True
            )
        else:
            to = Address(params["to"])
            if to in fqn_overrides:
                contract_fqn = fqn_overrides[to]
            else:
                contract_fqn = get_fqn_from_address(
                    Address(params["to"]), "latest", chain
                )
            if contract_fqn is None:
                logger.warning(f"Failed to get contract FQN for {params['to']}")

            self.process_trace(
                contract_fqn,
                debug_trace,
                chain,
                fqn_overrides,
                is_from_deployment=False,
            )

        if self._callback is not None:
            self._callback()

    def get_contract_ide_coverage(
        self,
    ) -> Dict[pathlib.Path, Dict[IdePosition, IdeFunctionCoverageRecord]]:
        """
        Returns coverage data for IDE usage
        """
        cov_data = {}
        for func, func_count in chain(
            self._function_coverage.items(), self._modifier_coverage.items()
        ):
            if func.file not in cov_data:
                cov_data[func.file] = {}

            func_ide_pos = IdePosition(
                *_get_line_col_from_offset(
                    func.name_location[0], self._lines_index[func.file]
                ),
                *_get_line_col_from_offset(
                    func.name_location[1], self._lines_index[func.file]
                ),
            )

            branch_records = {}

            for statement, count in self._statement_coverage.items():
                if statement.file != func.file:
                    continue
                if isinstance(
                    statement,
                    (DoWhileStatement, ForStatement, IfStatement, WhileStatement),
                ):
                    if statement.condition is None:
                        continue
                    start, end = statement.condition.byte_location
                elif isinstance(statement, TryStatement):
                    start, end = statement.external_call.byte_location
                elif isinstance(statement, (YulForLoop, YulIf)):
                    start, end = statement.condition.byte_location
                elif isinstance(statement, YulSwitch):
                    start, end = statement.expression.byte_location
                else:
                    start, end = statement.byte_location

                if start >= func.byte_location[0] and end <= func.byte_location[1]:
                    ide_pos = IdePosition(
                        *_get_line_col_from_offset(start, self._lines_index[func.file]),
                        *_get_line_col_from_offset(end, self._lines_index[func.file]),
                    )
                    branch_records[ide_pos] = IdeCoverageRecord(ide_pos, count)

            cov_data[func.file][func_ide_pos] = IdeFunctionCoverageRecord(
                name=func.name,
                ide_pos=func_ide_pos,
                coverage_hits=func_count,
                mod_records={},
                branch_records=branch_records,
            )

        return cov_data

    def process_trace(
        self,
        contract_fqn: Optional[str],
        trace: Dict[str, Any],
        chain: Chain,
        fqn_overrides: ChainMap[Address, Optional[str]],
        is_from_deployment: bool = False,
    ):
        """
        Processes debug_traceTransaction and it's struct_logs
        """
        contract_fqn_stack = [contract_fqn]
        is_deployment_stack = [is_from_deployment]
        assert len(fqn_overrides.maps) == 1

        for i, struct_log in enumerate(trace["structLogs"]):
            last_fqn = contract_fqn_stack[-1]
            deployment = is_deployment_stack[-1]
            pc = int(struct_log["pc"])

            if i > 0:
                prev_log = trace["structLogs"][i - 1]
                if (
                    prev_log["op"] in ("CALL", "CALLCODE", "DELEGATECALL", "STATICCALL")
                    and prev_log["depth"] == struct_log["depth"]
                ):
                    # precompiled contract was called in the previous trace
                    fqn_overrides.maps[1].update(fqn_overrides.maps[0])
                    fqn_overrides.maps.pop(0)
                    contract_fqn_stack.pop()
                    is_deployment_stack.pop()

            if struct_log["op"] in ("CALL", "CALLCODE", "DELEGATECALL", "STATICCALL"):
                logger.debug(f"Call {pc} {struct_log['op']} {struct_log['stack']}")
                addr = Address(int(struct_log["stack"][-2], 16))
                if addr in fqn_overrides:
                    new_fqn = fqn_overrides[addr]
                else:
                    new_fqn = get_fqn_from_address(addr, "latest", chain)
                contract_fqn_stack.append(new_fqn)
                is_deployment_stack.append(False)
                fqn_overrides.maps.insert(0, {})
            elif struct_log["op"] in ("CREATE", "CREATE2"):
                logger.debug(
                    f"Create call {pc} {struct_log['op']} {struct_log['memory']}"
                )
                offset = int(struct_log["stack"][-2], 16)
                length = int(struct_log["stack"][-3], 16)
                creation_code = read_from_memory(offset, length, struct_log["memory"])

                try:
                    new_fqn, _ = get_fqn_from_creation_code(creation_code)
                except ValueError:
                    logger.warning(f"Failed to get contract FQN for {creation_code}")
                    new_fqn = None

                contract_fqn_stack.append(new_fqn)
                is_deployment_stack.append(True)
                fqn_overrides.maps.insert(0, {})
            elif struct_log["op"] in {
                "INVALID",
                "RETURN",
                "STOP",
                "REVERT",
                "SELFDESTRUCT",
            }:
                logger.debug(f"{pc} {struct_log['op']} before pop {contract_fqn_stack}")
                if (
                    struct_log["op"] not in {"INVALID", "REVERT"}
                    and len(fqn_overrides.maps) > 1
                ):
                    fqn_overrides.maps[1].update(fqn_overrides.maps[0])
                fqn_overrides.maps.pop(0)

                if is_deployment_stack.pop():
                    try:
                        addr = Address(int(trace["structLogs"][i + 1]["stack"][-1], 16))
                        if addr != Address(0):
                            fqn_overrides.maps[0][addr] = contract_fqn_stack[-1]
                    except IndexError:
                        pass

                contract_fqn_stack.pop()

            pc_maps = self._pc_maps if not deployment else self._pc_maps_undeployed
            if last_fqn not in pc_maps:
                continue
            pc_map = pc_maps[last_fqn]

            if pc in pc_map:
                path = pc_map[pc].source_file
                if path is not None and path in self._interval_trees:
                    self._update_coverage(pc, pc_map[pc].offset, path)

        self._flush_coverage()

    def _update_coverage(
        self, pc: int, byte_offsets: Tuple[int, int], path: pathlib.Path
    ) -> None:
        interval_tree = self._interval_trees[path]
        start, end = byte_offsets
        intervals = interval_tree[start:end]
        nodes: List[IrAbc] = [interval.data for interval in intervals]

        functions = []
        modifiers = []
        statements = []
        yul_statements = []
        for node in nodes:
            if isinstance(node, FunctionDefinition):
                functions.append(node)
            elif isinstance(node, ModifierDefinition):
                modifiers.append(node)
            elif isinstance(node, StatementAbc):
                statements.append(node)
            elif isinstance(node, YulStatementAbc):
                yul_statements.append(node)

        function = None
        if len(functions) == 1:
            function = functions[0]
            if start >= function.byte_location[0] and end <= function.byte_location[1]:
                self._visited_functions.add(function)

        modifier = None
        if len(modifiers) == 1:
            modifier = modifiers[0]
            if start >= modifier.byte_location[0] and end <= modifier.byte_location[1]:
                self._visited_modifiers.add(modifier)

        if function is None and modifier is None:
            return
        decl = function if function is not None else modifier
        assert isinstance(decl, (FunctionDefinition, ModifierDefinition))

        if len(yul_statements) > 0:
            yul_statements.sort(key=lambda x: x.ast_tree_depth)
            yul_statement = yul_statements[-1]
            if isinstance(yul_statement, (YulBlock, YulFunctionDefinition)):
                return
            last_statement = self._last_statements[decl]
            if (
                last_statement is None
                or last_statement[0] != yul_statement
                or last_statement[1] >= pc
            ):
                if (
                    start >= yul_statement.byte_location[0]
                    and end <= yul_statement.byte_location[1]
                ):
                    self._statement_coverage[yul_statement] += 1
                    self._last_statements[decl] = (yul_statement, pc)
        elif len(statements) > 0:
            statements.sort(key=lambda x: x.ast_tree_depth)
            statement = statements[-1]
            if isinstance(statement, (Block, UncheckedBlock)):
                return
            last_statement = self._last_statements[decl]
            if (
                last_statement is None
                or last_statement[0] != statement
                or last_statement[1] >= pc
            ):
                if (
                    start >= statement.byte_location[0]
                    and end <= statement.byte_location[1]
                ):
                    self._statement_coverage[statement] += 1
                    self._last_statements[decl] = (statement, pc)

    def _flush_coverage(self) -> None:
        for fn in self._visited_functions:
            self._function_coverage[fn] += 1
        for mod in self._visited_modifiers:
            self._modifier_coverage[mod] += 1
        self._visited_functions.clear()
        self._visited_modifiers.clear()
        self._last_statements.clear()



from __future__ import annotations

from typing import TYPE_CHECKING, Union

from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import YulTypedName

from .abc import YulAbc

if TYPE_CHECKING:
    from .function_definition import FunctionDefinition
    from .variable_declaration import VariableDeclaration


class TypedName(YulAbc):
    """
    TBD
    """

    _parent: Union[FunctionDefinition, VariableDeclaration]
    _name: str
    _type: str

    def __init__(self, init: IrInitTuple, typed_name: YulTypedName, parent: YulAbc):
        super().__init__(init, typed_name, parent)
        self._name = typed_name.name
        self._type = typed_name.type

    @property
    def parent(self) -> Union[FunctionDefinition, VariableDeclaration]:
        return self._parent

    @property
    def name(self) -> str:
        return self._name

    @property
    def type(self) -> str:
        return self._type



from __future__ import annotations

import re
from functools import lru_cache
from typing import TYPE_CHECKING, Iterator, Optional, Tuple, Union

from Crypto.Hash import keccak

from .abc import DeclarationAbc

if TYPE_CHECKING:
    from .contract_definition import ContractDefinition

from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.ir.meta.parameter_list import ParameterList
from woke.ast.ir.meta.structured_documentation import StructuredDocumentation
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import SolcEventDefinition, SolcStructuredDocumentation


class EventDefinition(DeclarationAbc):
    """
    Definition of an event.
    !!! example
        ```solidity
        event Transfer(address indexed from, address indexed to, uint256 value);
        ```
    """

    _ast_node: SolcEventDefinition
    _parent: ContractDefinition

    _anonymous: bool
    _parameters: ParameterList
    _documentation: Optional[Union[StructuredDocumentation, str]]
    _event_selector: Optional[bytes]

    def __init__(
        self, init: IrInitTuple, event: SolcEventDefinition, parent: SolidityAbc
    ):
        super().__init__(init, event, parent)
        self._anonymous = event.anonymous
        self._parameters = ParameterList(init, event.parameters, self)

        if event.documentation is None:
            self._documentation = None
        elif isinstance(event.documentation, SolcStructuredDocumentation):
            self._documentation = StructuredDocumentation(
                init, event.documentation, self
            )
        elif isinstance(event.documentation, str):
            self._documentation = event.documentation
        else:
            raise TypeError(
                f"Unknown type of documentation: {type(event.documentation)}"
            )
        self._event_selector = (
            bytes.fromhex(event.event_selector) if event.event_selector else None
        )

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        yield from self._parameters
        if isinstance(self._documentation, StructuredDocumentation):
            yield from self._documentation

    def _parse_name_location(self) -> Tuple[int, int]:
        IDENTIFIER = r"[a-zA-Z$_][a-zA-Z0-9$_]*"
        EVENT_RE = re.compile(
            r"^\s*event\s+(?P<name>{identifier})".format(identifier=IDENTIFIER).encode(
                "utf-8"
            )
        )

        byte_start = self._ast_node.src.byte_offset
        match = EVENT_RE.match(self._source)
        assert match
        return byte_start + match.start("name"), byte_start + match.end("name")

    @property
    def parent(self) -> ContractDefinition:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def canonical_name(self) -> str:
        return f"{self._parent.canonical_name}.{self._name}"

    @property
    @lru_cache(maxsize=2048)
    def declaration_string(self) -> str:
        ret = (
            f"event {self._name}("
            + ", ".join(
                param.declaration_string for param in self.parameters.parameters
            )
            + f"){' anonymous' if self.anonymous else ''}"
        )
        if isinstance(self.documentation, StructuredDocumentation):
            return (
                "/// "
                + "\n///".join(line for line in self.documentation.text.splitlines())
                + "\n"
                + ret
            )
        elif isinstance(self.documentation, str):
            return (
                "/// "
                + "\n///".join(line for line in self.documentation.splitlines())
                + "\n"
                + ret
            )
        else:
            return ret

    @property
    def anonymous(self) -> bool:
        """
        !!! example
            ```solidity
            event Transfer(address indexed from, address indexed to, uint256 value) anonymous;
            ```
        Returns:
            `True` if the event is anonymous, `False` otherwise.
        """
        return self._anonymous

    @property
    def parameters(self) -> ParameterList:
        """
        Returns:
            Parameter list describing parameters of the event.
        """
        return self._parameters

    @property
    def documentation(self) -> Optional[Union[StructuredDocumentation, str]]:
        """
        Of [StructuredDocumentation][woke.ast.ir.meta.structured_documentation.StructuredDocumentation] type since Solidity 0.6.3.
        Returns:
            [NatSpec](https://docs.soliditylang.org/en/latest/natspec-format.html) documentation string, if any.
        """
        return self._documentation

    @property
    @lru_cache(maxsize=2048)
    def event_selector(self) -> bytes:
        """
        Returns:
            Selector of the event.
        """
        if self._event_selector is not None:
            return self._event_selector
        else:
            signature = f"{self._name}({','.join(param.type.abi_type() for param in self.parameters.parameters)})"
            h = keccak.new(data=signature.encode("utf-8"), digest_bits=256)
            return h.digest()



import logging
from typing import List, Optional, Union

import woke.ast.ir.yul as yul
from woke.ast.enums import ContractKind
from woke.ast.ir.abc import IrAbc
from woke.ast.ir.declaration.abc import DeclarationAbc
from woke.ast.ir.declaration.contract_definition import ContractDefinition
from woke.ast.ir.declaration.function_definition import FunctionDefinition
from woke.ast.ir.declaration.modifier_definition import ModifierDefinition
from woke.ast.ir.declaration.variable_declaration import VariableDeclaration
from woke.ast.ir.expression.identifier import Identifier
from woke.ast.ir.expression.member_access import MemberAccess
from woke.ast.ir.meta.identifier_path import IdentifierPath
from woke.ast.ir.type_name.user_defined_type_name import UserDefinedTypeName
from woke.ast.nodes import AstNodeId
from woke.lsp.common_structures import (
    DocumentUri,
    PartialResultParams,
    Range,
    StaticRegistrationOptions,
    SymbolKind,
    SymbolTag,
    TextDocumentPositionParams,
    TextDocumentRegistrationOptions,
    WorkDoneProgressOptions,
    WorkDoneProgressParams,
)
from woke.lsp.context import LspContext
from woke.lsp.lsp_data_model import LspModel
from woke.lsp.utils import path_to_uri, position_within_range, uri_to_path

logger = logging.getLogger(__name__)


class TypeHierarchyOptions(WorkDoneProgressOptions):
    pass


class TypeHierarchyRegistrationOptions(
    TextDocumentRegistrationOptions, TypeHierarchyOptions, StaticRegistrationOptions
):
    pass


class TypeHierarchyPrepareParams(TextDocumentPositionParams, WorkDoneProgressParams):
    pass


class TypeHierarchyItemData(LspModel):
    ast_node_id: int
    cu_hash: str
    uri: DocumentUri


class TypeHierarchyItem(LspModel):
    name: str
    """
    The name of this item.
    """
    kind: SymbolKind
    """
    The kind of this item.
    """
    tags: Optional[List[SymbolTag]]
    """
    Tags for this item.
    """
    detail: Optional[str]
    """
    More detail for this item, e.g. the signature of a function.
    """
    uri: DocumentUri
    """
    The resource identifier of this item.
    """
    range: Range
    """
    The range enclosing this symbol not including leading/trailing whitespace
    but everything else, e.g. comments and code.
    """
    selection_range: Range
    """
    The range that should be selected and revealed when this symbol is being
    picked, e.g. the name of a function. Muse be contained by the
    [`range`](#TypeHierarchyItem.range).
    """
    data: TypeHierarchyItemData
    """
    A data entry field that is preserved between a type hierarchy prepare and
    supertypes or subtypes requests. It could also be used to identify the
    type hierarchy in the server, helping improve the performance on
    resolving supertypes and subtypes.
    """


class TypeHierarchySupertypesParams(WorkDoneProgressParams, PartialResultParams):
    item: TypeHierarchyItem


class TypeHierarchySubtypesParams(WorkDoneProgressParams, PartialResultParams):
    item: TypeHierarchyItem


def _get_node_symbol_kind(
    node: Union[
        ContractDefinition, FunctionDefinition, ModifierDefinition, VariableDeclaration
    ]
) -> SymbolKind:
    if isinstance(node, ContractDefinition):
        if node.kind == ContractKind.CONTRACT:
            return SymbolKind.CLASS
        elif node.kind == ContractKind.INTERFACE:
            return SymbolKind.INTERFACE
        elif node.kind == ContractKind.LIBRARY:
            return SymbolKind.MODULE
        else:
            assert False, f"Unknown contract kind {node.kind}"
    elif isinstance(node, FunctionDefinition):
        if isinstance(node.parent, ContractDefinition):
            return SymbolKind.METHOD
        else:
            return SymbolKind.FUNCTION
    elif isinstance(node, ModifierDefinition):
        return SymbolKind.FUNCTION
    elif isinstance(node, VariableDeclaration):
        return SymbolKind.VARIABLE
    else:
        assert False, f"Unknown node type {type(node)}"


async def prepare_type_hierarchy(
    context: LspContext, params: TypeHierarchyPrepareParams
) -> Union[List[TypeHierarchyItem], None]:
    logger.debug(
        f"Type hierarchy for file {params.text_document.uri} at position {params.position} requested"
    )
    await context.compiler.output_ready.wait()

    path = uri_to_path(params.text_document.uri).resolve()

    if path not in context.compiler.interval_trees:
        return None

    tree = context.compiler.interval_trees[path]

    byte_offset = context.compiler.get_byte_offset_from_line_pos(
        path, params.position.line, params.position.character
    )
    intervals = tree.at(byte_offset)
    nodes: List[IrAbc] = [interval.data for interval in intervals]
    logger.debug(
        f"Found {len(nodes)} nodes at byte offset {byte_offset}:\n{sorted(nodes, key=lambda x: x.ast_tree_depth)}"
    )
    if len(nodes) == 0:
        return None

    node = max(nodes, key=lambda n: n.ast_tree_depth)
    logger.debug(f"Found node {node}")

    if isinstance(node, DeclarationAbc):
        name_location_range = context.compiler.get_range_from_byte_offsets(
            node.file, node.name_location
        )
        if not position_within_range(params.position, name_location_range):
            return None

    if isinstance(node, (Identifier, MemberAccess)):
        node = node.referenced_declaration
    elif isinstance(node, (IdentifierPath, UserDefinedTypeName)):
        part = node.identifier_path_part_at(byte_offset)
        if part is None:
            return None
        node = part.referenced_declaration
    elif isinstance(node, yul.Identifier):
        external_reference = node.external_reference
        if external_reference is None:
            return None
        node = external_reference.referenced_declaration

    if isinstance(
        node, (ContractDefinition, FunctionDefinition, ModifierDefinition)
    ) or (isinstance(node, VariableDeclaration) and node.overrides is not None):
        return [
            TypeHierarchyItem(
                name=node.canonical_name,
                kind=_get_node_symbol_kind(node),
                tags=None,
                detail=None,
                uri=DocumentUri(path_to_uri(node.file)),
                range=context.compiler.get_range_from_byte_offsets(
                    node.file, node.byte_location
                ),
                selection_range=context.compiler.get_range_from_byte_offsets(
                    node.file, node.name_location
                ),
                data=TypeHierarchyItemData(
                    ast_node_id=node.ast_node_id,
                    cu_hash=node.cu_hash.hex(),
                    uri=DocumentUri(path_to_uri(node.file)),
                ),
            )
        ]

    return None


async def supertypes(
    context: LspContext, params: TypeHierarchySupertypesParams
) -> Union[List[TypeHierarchyItem], None]:
    logger.debug(f"Supertypes for {params.item} requested")
    ast_node_id = params.item.data.ast_node_id
    cu_hash = bytes.fromhex(params.item.data.cu_hash)
    node = context.compiler.ir_reference_resolver.resolve_node(
        AstNodeId(ast_node_id), cu_hash
    )
    assert isinstance(
        node,
        (
            ContractDefinition,
            FunctionDefinition,
            ModifierDefinition,
            VariableDeclaration,
        ),
    )

    type_items = []
    nodes: List[
        Union[
            ContractDefinition,
            FunctionDefinition,
            ModifierDefinition,
            VariableDeclaration,
        ]
    ] = []
    if isinstance(node, ContractDefinition):
        for base_contract in node.base_contracts:
            contract = base_contract.base_name.referenced_declaration
            assert isinstance(contract, ContractDefinition)
            nodes.append(contract)
    elif isinstance(node, (FunctionDefinition, VariableDeclaration)):
        for base_function in node.base_functions:
            nodes.append(base_function)
    elif isinstance(node, ModifierDefinition):
        for base_modifier in node.base_modifiers:
            nodes.append(base_modifier)

    for node in nodes:
        type_items.append(
            TypeHierarchyItem(
                name=node.canonical_name,
                kind=_get_node_symbol_kind(node),
                tags=None,
                detail=None,
                uri=DocumentUri(path_to_uri(node.file)),
                range=context.compiler.get_range_from_byte_offsets(
                    node.file, node.byte_location
                ),
                selection_range=context.compiler.get_range_from_byte_offsets(
                    node.file, node.name_location
                ),
                data=TypeHierarchyItemData(
                    ast_node_id=node.ast_node_id,
                    cu_hash=node.cu_hash.hex(),
                    uri=DocumentUri(path_to_uri(node.file)),
                ),
            )
        )
    return type_items


async def subtypes(
    context: LspContext, params: TypeHierarchySubtypesParams
) -> Union[List[TypeHierarchyItem], None]:
    logger.debug(f"Subtypes for {params.item} requested")
    ast_node_id = params.item.data.ast_node_id
    cu_hash = bytes.fromhex(params.item.data.cu_hash)
    node = context.compiler.ir_reference_resolver.resolve_node(
        AstNodeId(ast_node_id), cu_hash
    )
    assert isinstance(
        node,
        (
            ContractDefinition,
            FunctionDefinition,
            ModifierDefinition,
            VariableDeclaration,
        ),
    )

    type_items = []
    nodes: List[
        Union[
            ContractDefinition,
            FunctionDefinition,
            ModifierDefinition,
            VariableDeclaration,
        ]
    ] = []
    if isinstance(node, ContractDefinition):
        for child_contract in node.child_contracts:
            nodes.append(child_contract)
    elif isinstance(node, FunctionDefinition) and node.child_functions:
        for child_function in node.child_functions:
            nodes.append(child_function)
    elif isinstance(node, ModifierDefinition):
        for child_modifier in node.child_modifiers:
            nodes.append(child_modifier)

    for node in nodes:
        type_items.append(
            TypeHierarchyItem(
                name=node.canonical_name,
                kind=_get_node_symbol_kind(node),
                tags=None,
                detail=None,
                uri=DocumentUri(path_to_uri(node.file)),
                range=context.compiler.get_range_from_byte_offsets(
                    node.file, node.byte_location
                ),
                selection_range=context.compiler.get_range_from_byte_offsets(
                    node.file, node.name_location
                ),
                data=TypeHierarchyItemData(
                    ast_node_id=node.ast_node_id,
                    cu_hash=node.cu_hash.hex(),
                    uri=DocumentUri(path_to_uri(node.file)),
                ),
            )
        )
    return type_items
